from rich.console import Console
console = Console()

class TranspilerExceptions:
    class NamespaceNotFound(Exception):
        def __init__(self, namespace, namespaces):
            Exception.__init__(self, f"No such namespace '{namespace}'. Available namespaces include: {list(namespaces.keys())}")
    class NotImplemented(Exception):
        def __init__(self):
            Exception.__init__(self, "Not implemented")
    class TypeMissmatch(Exception):
        def __init__(self, var, got, expected, line):
            Exception.__init__(self, f"Line {line}: For {var}, got: {got}, expected: {expected}")
    class TypeMissmatchNL(Exception):
        def __init__(self, var, got, expected):
            Exception.__init__(self, f"For {var}, got: {got}, expected: {expected}")
    class Overflow(Exception):
        def __init__(self, var, got, expected):
            Exception.__init__(self, f"For {var}, got: {got}, value overflows the given size of {expected}")
    class UnkownVar(Exception):
        def __init__(self, var, vars, nm="Nil"):
            Exception.__init__(self, f"Got unknown '{var}' (in {nm}). Variables in the current scope only contain: {list(vars.keys())}")
    class VarExists(Exception):
        def __init__(self, var):
            Exception.__init__(self, f"Got '{var}'. This variable already exists in this scope")
    class ClassExists(Exception):
        def __init__(self, cls, clss=[]):
            Exception.__init__(self, f"Got '{cls}'. This class already exists in this scope. {clss}")
    class ClassNotFound(Exception):
        def __init__(self, cls):
            Exception.__init__(self, f"Got '{cls}'. No such class exists within this scope")
    class InvalidIndexType(Exception):
        def __init__(self, type, types):
            Exception.__init__(self, f"Got type '{type}' for index. Supported types are: {types}")
    class OutOfBounds(Exception):
        def __init__(self, index, length):
            Exception.__init__(self, f"Index {index} is out of bounds. Array length is {length}")
    class UnauthorizedCall(Exception):
        def __init__(self, func, level, nm, clas):
            Exception.__init__(self, f"The function '{func}' is {level}, but being called from '{nm}', it's parent is '{clas}'")
    class UnauthorizedBitSet(Exception):
        def __init__(self):
            Exception.__init__(self, f"This program is not being built in the correct mode for setting the bit configuration.")
    class IncompleteDeclaration(Exception):
        def __init__(self, tree):
            Exception.__init__(self, f"The variable declaration does not contain any matched declaration method. Tree & constructor: {tree}")
    class UnknownActionReference(Exception):
        def __init__(self, act):
            Exception.__init__(self, f"Got unknown action '{act}'")
    class UnknownMode(Exception):
        def __init__(self, mode):
            Exception.__init__(self, f"Got unknown action mode '{mode}'")
    class UnkownMethodReference(Exception):
        def __init__(self, func, funcs, ns='Nil'):
            Exception.__init__(self, f"Got unknown function '{func}' ({ns}), from '{funcs}'")
    class ActionRequiredButNotFound(Exception):
        def __init__(self, act, acts):
            Exception.__init__(self, f"An action '{act}' is required for this task. But only {list(acts.keys())} are installed")
    class UnkownType(Exception):
        def __init__(self, type):
            Exception.__init__(self, f"Unknown type supplied {type}")
    class TooManyValues(Exception):
        def __init__(self, values, scope):
            Exception.__init__(self, f"Too many values supplied to unpack in current scope. values: {values} scope: {scope}")
    class IncorrectNumberOfValues(Exception):
        def __init__(self, values, scope):
            Exception.__init__(self, f"Invalid number of values: {values} scope: {scope}")
    class Generic(Exception):
        def __init__(self, msg):
            Exception.__init__(self, msg)

class RuntimeExceptions:
    class Panic(Exception):
        def __init__(self, msg, line="Unknown"):
            Exception.__init__(self, f"at line {line} | {msg}")

class ModuleExceptions:
    class InvalidModuleConstruction(Exception):
        def __init__(self, module):
            Exception.__init__(self, f"[Module:{module.name}] Invalid arguments supplied to constructor")


def Warn(msg):
    print("Warning", end="")
    console.log(f"[red]{msg}[/red]")



from ..compiler import Compiler
from ..exceptions import TranspilerExceptions
from ..module import Module
from ..types import *
import json

class EchoMod(Module):
    name="echo"
    type = Module.MODULE_TYPES.SPECIAL_ACTION_FUNC

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree, stubborn=False):
        # Dependencies
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')
        values = self.base(tree, ref=stubborn)

        if len(values) > 1:
            raise TranspilerExceptions.TooManyValues(values, "echo($msg)")
        
        value = values[0]

        # Patch echo to correctly handle JSON

        if type(value) in (list, dict, tuple):
            value = json.dumps(value)
        
        # print(value)

        return str(value)






from .exceptions import TranspilerExceptions
from .modules import Module
from .types import Nil
from rich import print as rprint
from copy import deepcopy

from rich.console import Console
from rich.syntax import Syntax
console = Console()

class Compiler:
    def __init__(
            self,
            tree,
            actions = {},
            variables = {},
            functions = {},
            classes = {},
            builtin_functions = {},
            namespace = "main",
            namespaces = {}
    ):
        self.code = tree
        self.actions = actions 
        self._modules = {}
        self.line = 0
        self.variables: dict = variables
        self.builtin_functions = {}
        self.functions = functions
        self.namespace = namespace
        self.namespaces = namespaces
        self.classes = classes
        self.finished = []
        self.finished_funcs = []

        self.parent = None

        self.returns = Nil(None) # No return
        self.stop = False

        # Keep track of current and previous module run
        self.curr = None
        self.prev = None

        self.persistent = False

    def new_instance(
        self,
        tree = None,
        namespace = '_anon',
        namespaces = {},
        variables = {},
        functions = {},
        classes = {},
        sync = ""
    ):
        # Modules and actions need to be recreated with a new scope
        instance = Compiler(
            tree = tree,
            namespace = namespace,
            variables = self.variables if sync in ("object", "advanced") else self.deepcopy_vars(),
            functions = self.functions.copy() if sync == "advanced" else {},
            classes = self.classes if sync in ("class", "object", "advanced") else {},
            namespaces = namespaces,
        )

        Compiler.populate_modules_actions(
            instance,
            self.raw_modules
        )

        # builtins should reflect their new namespace
        for builtin in self.builtin_functions:
            instance.builtin_functions[builtin] = {
                "run_func": type(self.builtin_functions[builtin]['run_func'])(instance)
            }

        instance.parent = self

        return instance
    
    def deepcopy_vars(self):
        v = {}
        for var in self.variables:
            try:
                v[var] = deepcopy(self.variables[var])
            except:
                print(f"Was unable to clone : {var}")
        return v
    
    def __deepcopy__(self, memodict={}):
        copy_object = Compiler(
            tree=[]
        )
        copy_object.functions = self.functions
        return copy_object
    
    def sync_functions(
        self,
        instance
    ):
        for name in instance.functions:
            self.functions[name] = instance.functions[name]
            

    def sync_variables(self, instance):
        for name in instance.variables:
            self.variables[name] = instance.variables[name]
            

    def sync_classes(self, instance: 'Compiler'):
        for name in instance.classes:
            self.classes[name] = instance.classes[name]

    def sync(self, instance: 'Compiler'):
        self.sync_functions(instance)
        self.sync_variables(instance)
        self.sync_classes(instance)

    
    def populate_modules_actions(
        self,
        Modules: list[Module]
    ) -> None:
        # Issue 1: 19-12-23
        #
        # These are fucked...
        # For some reason, they affect the parent compiler instance
        # This should not happen!?
        # I have no idea why this happens!?
        # Hack 1 fixes this... But it's a bad idea
        #   Update: 20-12-23
        # The issue persists with subclass instances
        # This is stupid
        # I have no idea whats causing this
        #   Update: 23-12-23
        # I've been able to track down the id's of these instances
        # They seem to be the same even after 
        # a rebuild of the parent's source
        # Will try to run the rebuild after the sync
        # Turns out this is done by default.
        # I'm still unsure about how to fix the issue
        # Hoping that pycharms debugging features can save my ass
        #   Update: 24-12-23
        # Still no fix
        # Update: 24-12-23
        # Issue has been fixed, the issue was in how the objects were being created.
        # NEVER use deep copy for complex structures.....
        # Update: 25-12-23
        # The issue was not fixed... Scopes are still broken
        # Update 26-12-23
        # The issue has been resolved :pain: hopefully for real this time
        self.raw_modules = Modules
        for module in Modules:
            modobj: Module = module(self)
            modobj.type = module.type
            if modobj.type == Module.MODULE_TYPES.FUNCTION:
                self.functions[module.name] = modobj
            else:
                self.actions[module.name] = modobj

            if module.type == Module.MODULE_TYPES.SPECIAL_ACTION_FUNC:
                self.functions[module.name] = modobj

            self._modules[module.name] = modobj

    def add_builtin_functions(self, funcs):
        self.builtin_functions = funcs

        # Add them as actions too
        for func in funcs:
            self.actions[funcs[func]["run_func"].name] = funcs[func]["run_func"]
    
    def get_action(
        self,
        name
    ) -> Module:
        if name in self.actions:
            return self.actions[name]
        else:
            print(f"At {self.line}")
            raise TranspilerExceptions.ActionRequiredButNotFound(name, self.actions)

    def run(self, code=None):
        # Fixed: Ok so, the issue here was that sometimes the function body wouldnt be ran properly
        # This was due to self.stop persisting. The below disables that 
        self.stop = False
        if code is None:
            code = self.code
        else:
            code = code
        for action in code:
            if self.stop:
                return
            # Hack 1, see issue 1
            self.populate_modules_actions(
                self.raw_modules
            )
            # print(self.namespace, self.variables)
            if action[0] in self.actions:
                # print(self.namespace, "->", action[0])
                self.curr = action[0]
                try:
                    ret = self.actions[action[0]](action[1])
                
                    if self.actions[action[0]].type == Module.MODULE_TYPES.SPECIAL_ACTION:
                        if ret != None:
                            if 'ID' in action[1] and type(action[1]['ID']) == tuple and\
                                self.get_action("RESOLUT")(action[1]['ID']).value == 'echo':
                                self.finished.append(ret)
                    elif self.actions[action[0]].type == Module.MODULE_TYPES.SPECIAL_ACTION_FUNC:
                        if ret != None:
                            self.finished_funcs.append(ret)
                    elif self.actions[action[0]].type == Module.MODULE_TYPES.NON_WRITEABLE:
                        continue
                except Exception as e:
                    console.print_exception(show_locals=False)
                    console.bell()
                    console.print(f"[red]({__file__.split('/')[-1]}:{action[-1]})\n\t({self.namespace})ERROR:[/red]", e)
                    rprint(
                        Syntax(
                            str(action) + '\nPrevious:\t'+ str(self.prev),
                            "lisp",
                            padding=1, 
                            line_numbers=True
                        )
                    )
                    print("Stacktrace:")
                    parent = self.parent
                    while parent is not None:
                        print(parent)
                        parent = parent.parent
                    if self.persistent:
                        raise e
                    else:
                        exit(1)
            else:
                print(code)
                raise TranspilerExceptions.UnknownActionReference(action[0])
            self.line += 1
            self.prev = action[0]

    def pause(self):
        rprint("[yellow bold]Compilation paused. Press enter to continue[/yellow bold]")
        input()

    def get_variable(
        self,
        name: str
    ) -> dict:
        if name in self.variables:
            return self.variables[name]
        else:
            print(f"At {self.line} {self.namespace} during {self.curr}")
            raise TranspilerExceptions.UnkownVar(name, self.variables, nm=self.namespace)

    def create_variable(
        self,
        name: str, 
        type,
        obj,
        level = 'public',
        force = False
    ) -> None:
        # print(f"At {self.line} {self.namespace}")
        if name in self.variables and not force:
            print(f"At {self.line}")
            raise TranspilerExceptions.VarExists(name)
        self.variables[name] = {
                "type": type,
                "object": obj,
                "level": level,
            }
        
    def remove_variable(
        self,
        name: str
    ) -> None:
        if name in self.variables:
            del self.variables[name]
        else:
            raise TranspilerExceptions.UnkownVar(name, self.variables)
        
    def create_function(
        self,
        name: str, 
        obj: object,
        source: str,
        returns,
        arguments=[],
        level = "public",
    ) -> None:
        self.functions[name] = {
                "object": obj,
                "source": source,
                "run_func": obj.run,
                "returns": returns,
                "arguments": arguments,
                "level": level
            }
        
    def write_back(self, origin: 'Compiler'):
        console.log(f"[red]<===>[/red] Writing back from '{origin.namespace}' to '{self.namespace}'")
        self.finished = self.finished + origin.finished
        self.finished_funcs = self.finished_funcs + origin.finished_funcs
        
    def create_class(
        self,
        name: str = '',
        obj = '',
        source = '',
        parent = None,
        force = False
    ) -> None:
        if name in self.classes and not force:
            print(f"At {self.line}")
            raise TranspilerExceptions.ClassExists(name, clss=self.classes)
        
        obj.parent = parent
        
        self.classes[name] = {
                "object": obj,
                "source": source,
                "parent": parent
            }
        
    
    def __str__(self) -> str:
        return f"NS::{self.namespace}"
    
    def __repr__(self) -> str:
        return self.__str__()
    
    def clean(self):
        self = Compiler([])
        self.variables = {}
        self.functions = {}
        self.classes = {}
        self.finished = []
        self.code = []
        self.namespaces = {}
        self.namespace = 'main'
        self.parent = None
        return self



import re


def split_php(text):
    pattern = r'(<\?php.*?\?>)|(.*)'

    # Find all matches
    matches = re.findall(pattern, text, re.DOTALL)

    # Extract the portions inside and outside the PHP code block
    text = ''.join(match[0] for match in matches).strip()
    html = ''.join(match[1] for match in matches).strip()
    return text, html



from ..exceptions import TranspilerExceptions, Warn
from ..module import Module
from ..types import *


class CommonMod(Module):
    name="COMMON"
    type = Module.MODULE_TYPES.FUNCTION

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def base(self, tree, ref=False):
        values = []

        if 'FUNCTION_ARGUMENTS' not in tree:
            # Advanced handling
            for var in tree:
                value = self.ref_resolve(var) if ref else self.safely_resolve(var)
                values.append(value)
        else:
            if 'POSITIONAL_ARGS' in tree['FUNCTION_ARGUMENTS']:
                for var in tree['FUNCTION_ARGUMENTS']['POSITIONAL_ARGS']:

                    value = self.ref_resolve(var) if ref else self.safely_resolve(var)

                    values.append(value)

        return values
    


# count($value) -> len($value)
class CountMod(CommonMod):
    name="count"
    type = Module.MODULE_TYPES.FUNCTION

    def __init__(self, compiler_instance):
        super().__init__(compiler_instance)
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        values = self.base(tree)

        return Int32(len(values[0]))
    
# empty($value)
class EmptyMod(CommonMod):
    name="empty"
    type = Module.MODULE_TYPES.FUNCTION

    def __init__(self, compiler_instance):
        super().__init__(compiler_instance)
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        value = self.base(tree)[0]

        emptiness = value is None or value == "" or value == [] or value == {} or value == set()

        return Bool(emptiness)

# For arrays
# array()
class ArrayMod(CommonMod):
    name="array"
    type = Module.MODULE_TYPES.FUNCTION

    def __init__(self, compiler_instance):
        super().__init__(compiler_instance)
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        Warn("It is discouraged to use array(), instead construct an empty array via []")
        return DynArray([])


# array_push($array, $value)
class ArrayPushMod(CommonMod):
    name="array_push"
    type = Module.MODULE_TYPES.FUNCTION

    def __init__(self, compiler_instance):
        super().__init__(compiler_instance)
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        array, value = self.base(tree, ref=True)[:2]

        array.value.append(value)

        return Bool(True)
    
_MODS = {
    "count": CountMod,
    "ArrayMod": ArrayMod,
    "array_push": ArrayPushMod,
    "empty": EmptyMod
}

def build_funcs(c):
    functions = {}
    for f in _MODS:
        functions[f] = {
                "run_func": _MODS[f](c)
            }
    return functions


import base64
from ..exceptions import TranspilerExceptions, Warn
from ..module import Module
from ..types import *
from ..compiler import Compiler
from json import dumps
from flask import jsonify

class CommonJSONMod(Module):
    name="COMMONJSON"
    type = Module.MODULE_TYPES.FUNCTION

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def base(self, tree, ref=False):
        values = []

        if 'FUNCTION_ARGUMENTS' not in tree:
            # Advanced handling
            for var in tree:
                value = self.ref_resolve(var) if ref else self.safely_resolve(var)
                values.append(value)
        else:
            if 'POSITIONAL_ARGS' in tree['FUNCTION_ARGUMENTS']:
                for var in tree['FUNCTION_ARGUMENTS']['POSITIONAL_ARGS']:

                    value = self.ref_resolve(var) if ref else self.safely_resolve(var)

                    values.append(value)

        return values


# json_encode($value) -> Map($value.__attrs)
class JSONEncodeMod(CommonJSONMod):
    name="json_encode"
    type = Module.MODULE_TYPES.FUNCTION

    def __init__(self, compiler_instance):
        super().__init__(compiler_instance)
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        values = self.base(tree)
        # We optimally want to extract the attributes of the 
        # object which will be value[0].
        # An object (Compiler obj) has the local variables
        # stored in self.variables, which makes it less than
        # optimal to get the attributes. Due to this, we will
        # grab all it's functions that are in the form of "get<N?>"
        # i.e. getters.
        attrs = {}
        obj = values[0] 
        if type(obj) == Compiler:
            for func in obj.functions:
                if func[:3].lower() == 'get':
                    # add their return values to our tracker
                    obj.functions[func]["run_func"]() 
                    attrs[func[3:].lower()] = obj.functions[func]["object"].returns
            return Map(attrs)
        elif type(obj) == list:
            return dumps(obj)
        elif type(obj) == dict:
            return dumps(obj)
        else:
            raise TranspilerExceptions.Generic(f"Couldn't convert {obj} to JSON")

    
_MODS = {
    "json_encode": JSONEncodeMod,
}

def build_funcs(c):
    functions = {}
    for f in _MODS:
        functions[f] = {
                "run_func": _MODS[f](c)
            }
    return functions


# Pretty-print
from rich import print
from rich.console import Console

from .lang.compiler import Compiler
from .lang.std.htmlspecialchars import HTMLSpecialCharMod
from .lang.std.rand import RandMod
from .lang.std.string import *
from .splitter import split_php
console = Console()
from rich.syntax import Syntax
from rich.progress import Progress

# Parser and lexer
from .lang.lexer import PyettyLexer
from .lang.pparser import PyettyParser
# from rich.traceback import install
# install()

# Load modules
from .lang.modules import (
    ConditionalMod,
    ForEachMod,
    GetIndexMod,
    InternalMod,
    MathMod,
    Module,
    NamespaceMemberMod,
    NamespaceMod,
    NewObjectMod,
    PhpMod,
    HTMLMod,
    ConcatMod,
    PublicMod,
    RequireOnceMod,
    ReturnMod,
    TarrowMod,
    UseMod,
    VariableAssignMod,
    ResolutionMod,
    FunctionCallMod,
    ClassDeclarationMod,
    FunctionDecMod,
    WhileMod
)
# Language builtins
from .lang.std.echo import EchoMod

### Args proc ###
from sys import argv
flags = ''
if len(argv) < 2: 
    raise Exception("Usage: python3 p <input>")
#################
### Read file ###
text = open(argv[1],"r").read()
text, html = split_php(text)
#################

### Construct tree ###
lexer = PyettyLexer()
parser = PyettyParser()
######################
def prettify_and_write(data, indentation=0, file=None):
    if isinstance(data, tuple):
        if len(data) == 2:
            label, content = data
            if isinstance(content, dict):
                file.write(f"{' ' * indentation}('{label}', {{\n")
                for key, value in content.items():
                    file.write(f"{' ' * (indentation + 4)}'{key}': ")
                    prettify_and_write(value, indentation + 4, file)
                file.write(f"{' ' * indentation}}},\n")
            elif isinstance(content, tuple):
                file.write(f"{' ' * indentation}('{label}', (\n")
                for item in content:
                    prettify_and_write(item, indentation + 4, file)
                file.write(f"{' ' * indentation})),\n")
            else:
                file.write(f"{' ' * indentation}('{label}', {content}),\n")
        elif len(data) == 1:
            prettify_and_write(data[0], indentation, file)
        else:
            for item in data:
                prettify_and_write(item, indentation + 4, file)
            
    elif isinstance(data, dict):
        file.write(f"{' ' * indentation}{{\n")
        for key, value in data.items():
            file.write(f"{' ' * (indentation + 4)}'{key}': ")
            prettify_and_write(value, 1, file)
        file.write(f"{' ' * indentation}}},\n")
    else:
        file.write(f"{' ' * indentation}{data},\n")




### Tasks ###

# Works only on functions that support tasks
def construct_task(
    progress, 
    task,
    target,
    *args,
    tasklen = 100
):
    def _run_task():
        return target(
            *args,
            task = task,
            progress = progress,
            tasklen = tasklen
        )
    return _run_task

with Progress(
    refresh_per_second = 100
) as total_progress:
    # Lex 
    lex = total_progress.add_task("Lex...", total=100)
    lexed_tokens = lexer.tokenize(text)

    # Increases time it takes, but makes it possible to track tree building
    lexed_len    = len(list(lexer.tokenize(text)))
    total_progress.update(lex, advance=100)

    # print(list(lexer.tokenize(text)))

    new_arr = []
    for tok in lexer.tokenize(text):
        new_arr.append(tok)
    


    # print(new_arr)

    # Build tree task
    build_tree = total_progress.add_task("Building tree...", total=1000)
    build_tree_task = construct_task(
        total_progress,
        build_tree,
        parser.parse,
        iter(new_arr),
        tasklen = len(new_arr)
    )

    tree = build_tree_task()
    with open("tree", "w+") as f:
        prettify_and_write(tree, 1, f)
    
    # Force finish
    total_progress.update(build_tree, advance=1000)

print(f"\n[bold green]Tree complete, translating...[/bold green]")

compiler = Compiler(tree)
compiler.populate_modules_actions(
    [
        PhpMod,
        HTMLMod,
        VariableAssignMod,
        ResolutionMod,
        NamespaceMod,
        FunctionCallMod,
        ConcatMod,
        ClassDeclarationMod,
        FunctionDecMod,
        ReturnMod,
        NewObjectMod,
        ConditionalMod,
        WhileMod,
        MathMod,
        NamespaceMemberMod,
        GetIndexMod,
        ForEachMod,
        InternalMod,
        TarrowMod
    ]
)
compiler.builtin_functions = {
    # Fake internals as functions
    "echo": {
        "run_func": EchoMod(compiler)
    },
    "require_once": {
        "run_func": RequireOnceMod(compiler)
    },
    "use": {
        "run_func": UseMod(compiler)
    },
    "public": {
        "run_func": PublicMod(compiler)
    },
    "rand": {
        "run_func": RandMod(compiler)
    },
    "htmlspecialchars": {
        "run_func": HTMLSpecialCharMod(compiler)
    },
    "strlen": {
        "run_func": StrLenMod(compiler)
    },
    "str_replace": {
        "run_func": StrReplaceMod(compiler)
    },
}

def main():
    compiler.run()

    pyetty_out_pretty = Syntax(
                        '\n'.join([str(el) for el in compiler.finished]) + "\n" + html,
                        "v",
                        padding=1, 
                        line_numbers=True
                    )
    # print(pyetty_out_pretty)

    print("[green]Complete serving...[/green]")

    with open('out.html', "w+") as f:
        f.write(''.join([str(el) for el in compiler.finished]))
        f.write(html)

if __name__ == '__main__':
    main()

# webbrowser.open('file://' + os.path.realpath("out.html"))

# print(f"\n[bold green]Source complete, compiling...[/bold green]")
# with open("out.v", "w+") as out:
#     out.write('\n'.join(compiler.finished))
# os.system(f"v crun out.v")


import json
from werkzeug.datastructures import ImmutableMultiDict

class BasicType:
    def __init__(self):
        self.value: object
    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.value})"
    def __repr__(self) -> str:
        return self.__str__()
    
    def __len__(self) -> int:
        return len(self.value)
    
    def __html__(self):
        return str(self.value)

class String(BasicType):
    '''
    Simple String type
    '''
    def __init__(self, value):
        self.value = f'"{value}"'
        self.length = len(value)
        super().__init__()

class ID(BasicType):
    '''
    Simple ID type
    '''
    def __init__(self, value):
        self.value = value
        self.length = len(value)
        super().__init__()

class Char(BasicType):
    '''
    Simple Char type
    '''
    size = 1
    abbr_name = 'char'
    def __init__(self, value):
        self.value = ord(value) if type(value) == str else value
        self.length = value
        self.size = Char.size
        self.hex  = False
        super().__init__()

class Bool(BasicType):
    '''
    Simple Bool type
    '''
    size = 1
    abbr_name = 'bool'
    def __init__(self, value):
        self.value = True if value == 'true' or value == True else False
        self.hex  = False
        super().__init__()

class Int32(BasicType):
    '''
    Simple Int32 type
    '''
    size = 4
    abbr_name = 'int'
    def __init__(self, value):
        self.value = int(value)
        self.length = value
        self.size = Int32.size
        self.hex = False
        super().__init__()

class Float(BasicType):
    '''
    Simple float type
    '''
    size = 4
    abbr_name = 'float'
    def __init__(self, value):
        self.value = float(value)
        self.length = value
        self.size = Int32.size
        self.hex = False
        super().__init__()

class HexInt32(BasicType):
    '''
    Simple Int32 type
    '''
    size = 4
    abbr_name = 'hex'
    def __init__(self, value):
        self.value = value
        self.length = len(value)
        self.size = Int32.size
        self.hex = True
        super().__init__()

class List(BasicType):
    '''
    Simple list type
    '''
    size = 0
    abbr_name = 'list'
    def __init__(self, values, type):
        self.value = values
        self.size = List.size
        self.offset = 0
        self.type = type
        super().__init__()

    @property
    def length(self):
        return len(self.value)

    # Types
    BONE = "Barebone"
    GUESS = "Guess it!"

class Map(BasicType):
    '''
    Simple map type
    '''
    size = 0
    abbr_name = 'Map'
    def __init__(self, values):
        self.value = values
        self.length = len(list(values.keys()))
        self.size = Map.size
        super().__init__()

# TODO: Discourage array() construction in favour of []

class DynArray(List):
    '''
    A dynamic size array
    '''
    size = 0
    abbr_name = 'DynamicArray'
    def __init__(self, values, type=''):
        super().__init__(values, type)
        self.value = list(values)
        # This should adapt to be an ASSOC Array if needed, 
        # or we could have another sub-type?
        # idfk

class Nil(BasicType):
    '''
    A nil or None equivalent 
    '''
    size = 0
    abbr_name = 'Nil'
    value = 'Nil'
    def __init__(self, _=''):
        self.value = ""
        super().__init__()

class Any(BasicType):
    '''
    Any, guess type later
    '''
    size = 0
    abbr_name = 'Any'
    def __init__(self, _):
        self.value = ""
        super().__init__()

class Auto(BasicType):
    '''
    This type will automatically fill 
    in the duties of other types.

    This is used for wrapping Python objects.

    For example: 

        Basic matching.

        We expect that the object contains all or most
        of the known types
        ```
        [Request]      -->      [Auto(Request)]
        |- str(method)            |- String(method)
        |- bytes(data)            |- DynArray(data, type=Char)
        |- ...                    |- ...
        ```
        
        Special matching.

        We expect a manually defined AutoEncoder and
        AutoDecoder for this object
    '''
    size = 0
    abbr_name = 'Auto'
    def __init__(self, v, type='none'):
        self.value = self.match_value(v, type) 
        self.match_type = type
        self.length = -1
        super().__init__()

    def match(self):
        "Dangerous, instead expect Auto to resolve automatically"
        return self.match_value(self.value, 'basic')

    def match_value(self, value, _type='none'):
        if type(value) == Auto:
            value = value.value
        if _type == 'basic':
            _f = {
                None: Nil,
                str: String,
                int: Int32,
                dict: Map,
                list: DynArray,
                tuple: DynArray,
                # bad bad bad
                type(None): Nil,
                ImmutableMultiDict: Map
            }
            if type(value) in _f:
                return _f[type(value)](value)
            return value
        else:
            return value

class sInnerMut(BasicType):
    '''
    System Internal Type
    '''
    size = 0
    abbr_name = 'sINNER'
    def __init__(self, *v):
        self.value = v
        super().__init__()


class Session:
    '''
    Session type.
    This is a special type variant as we need to access Flask's
    internal session object.
    '''
    size = 0
    abbr_name = 'Session'
    def __init__(self):
        pass

    @property
    def value(self):
        from flask import session
        return session
    

class Request:
    '''
    Request type.
    This is a special type variant as we need to access Flask's
    internal request object.
    '''
    size = 0
    abbr_name = 'Request'
    def __init__(self):
        pass

    @property
    def value(self):
        from flask import request
        return request
    
class SqlConnector(BasicType):
    '''
    SQL connection wrapper object 
    '''
    size = 0
    abbr_name = 'Connection'
    def __init__(self, connection):
        self.value = connection
        super().__init__()


class Salt(BasicType):
    '''
    Salt type 
    '''
    size = 0
    abbr_name = 'Salt'
    def __init__(self, salt):
        self.value = salt
        super().__init__()

NOPHP_TYPES = (
    String,
    Int32,
    ID,
    Char,
    Bool,
    Float,
    List,
    Map,
    DynArray,
    Auto,
    Any,
    Nil,
    sInnerMut,
    Session,
    Request,
    SqlConnector,
    Salt
)

BASE_TYPES = (
    str,
    list,
    int,
    tuple,
    float
)


import base64
from ..exceptions import TranspilerExceptions, Warn
from ..module import Module
from ..types import *


class CommonSaltMod(Module):
    name="COMMONSALT"
    type = Module.MODULE_TYPES.FUNCTION

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def base(self, tree, ref=False):
        values = []

        if 'FUNCTION_ARGUMENTS' not in tree:
            # Advanced handling
            for var in tree:
                value = self.ref_resolve(var) if ref else self.safely_resolve(var)
                values.append(value)
        else:
            if 'POSITIONAL_ARGS' in tree['FUNCTION_ARGUMENTS']:
                for var in tree['FUNCTION_ARGUMENTS']['POSITIONAL_ARGS']:

                    value = self.ref_resolve(var) if ref else self.safely_resolve(var)

                    values.append(value)

        return values


# gensalt($value) -> Salt
class GenSaltMod(CommonSaltMod):
    name="gensalt"
    type = Module.MODULE_TYPES.FUNCTION

    def __init__(self, compiler_instance):
        super().__init__(compiler_instance)
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        from bcrypt import gensalt

        values = self.base(tree)

        slowness = values[0]

        salt = gensalt(slowness)

        return Salt(salt)
    
# hashpw($value) -> String
class HashPWMod(CommonSaltMod):
    name="hashpw"
    type = Module.MODULE_TYPES.FUNCTION

    def __init__(self, compiler_instance):
        super().__init__(compiler_instance)
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        from bcrypt import hashpw

        values = self.base(tree)
        text, salt = values[0], values[1]
        enc = hashpw(text.encode('utf-8'), salt)

        # Encode into a string
        string = base64.b64encode(enc).decode('utf-8')

        return String(string)
    
# checkpw($value) -> Bool
class CheckPWMod(CommonSaltMod):
    name="checkpw"
    type = Module.MODULE_TYPES.FUNCTION

    def __init__(self, compiler_instance):
        super().__init__(compiler_instance)
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        from bcrypt import checkpw, hashpw, gensalt

        values = self.base(tree)
        text, hashed = values[0], values[1]


        # To bytes
        hashed = base64.b64decode(hashed.encode('utf-8'))
        valid = checkpw(text.encode('utf-8'), hashed)

        return Bool(valid)
    
_MODS = {
    "bcrypt_gensalt": GenSaltMod,
    "bcrypt_hashpw": HashPWMod,
    "bcrypt_checkpw": CheckPWMod
}

def build_funcs(c):
    functions = {}
    for f in _MODS:
        functions[f] = {
                "run_func": _MODS[f](c)
            }
    return functions


import markupsafe
from .types import *
from .exceptions import ModuleExceptions


class Module:
    BUILT_TYPE = str
    class MODULE_TYPES:
        '''
        FUNCTION -> Function module type
        ACTION   -> Internal action used in processing
        '''
        FUNCTION = "Function" 
        ACTION   = "Action"
        SPECIAL_ACTION = "Special Action"
        SPECIAL_ACTION_FUNC = "Special Action Function"
        NON_WRITEABLE = "Something that shouldnt need to be written to the output"

    def __init__(
        self,
        #type: MODULE_TYPES
    ):
        self.type = "Unknown"
        self.built = ""
        self.template = ""
        self.o1 = False
        self.o2 = False
        self.o3 = False
        self.no_construct = False        

    def __call__(
        self,
        tree,
        no_construct = False
    ):

        self.no_construct = no_construct
        _values = self.proc_tree(tree)
        self.override()
        return _values
    
    # Future
    #def optimise(self): pass
    optimise = None

    def remove_quotes(_,s):
        if type(s) != str: return str(s)
        if len(s) >= 2 and s[0] == '"' and s[-1] == '"':
            return s.strip('"')
    
        return s

    # Future: Return true for now
    def verify(self): return True

    def override(self):
        """
        Warning: this should be used with caution since it bypasses the default build structure
        """
        pass

    # Implemented in module child
    def proc_tree(self, tree) -> dict: "Return a dict containing values processed from the tree"

    # Format
    # TODO: Remove this as it's not used
    def _constructor(
        self,
        arguments: dict
    ) -> BUILT_TYPE:
        if arguments == None:
            raise ModuleExceptions.InvalidModuleConstruction(self)
        try:
            return self.template.format(
                    **arguments
                )
        except Exception:
            raise Exception(f"In '{self.name}' - Failed to unpack elements. Perhaps you need to set `no_construct` to True to avoid this module's construction?")

    def base(self, tree, ref=False):
        values = []

        if 'FUNCTION_ARGUMENTS' not in tree:
            # Advanced handling
            for var in tree:
                value = self.ref_resolve(var) if ref else self.safely_resolve(var)
                values.append(value)
        else:
            if 'POSITIONAL_ARGS' in tree['FUNCTION_ARGUMENTS']:
                for var in tree['FUNCTION_ARGUMENTS']['POSITIONAL_ARGS']:

                    value = self.ref_resolve(var) if ref else self.safely_resolve(var)

                    values.append(value)

        return values

    def safely_resolve(self, var):
        # Dependencies
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')
        func_module: Module = self.compiler_instance.get_action('FUNCTION_CALL')
        if type(var) == tuple:
            resolved = resolution_module(var)
        else: resolved = var
        value: BasicType = None

        # print(resolved)

        if type(resolved) == Auto:
            resolved = resolved.match()

        if type(resolved) == ID:
            value = resolved.value
            v = self.compiler_instance.get_variable(value)
            # print(v)
            value = self.safely_resolve(v['object'])
        elif type(resolved) == String:
            value = self.remove_quotes(resolved.value)
        elif type(resolved) == Int32:
            value = resolved.value
        elif type(resolved) == Bool:
            value = resolved.value
        elif type(resolved) == sInnerMut:
            value = func_module.run_sInnerMut(resolved).value
        elif type(resolved) == type(self.compiler_instance):
            value = resolved
        elif type(resolved) == SqlConnector:
            value = resolved.value
        elif type(resolved) == Salt:
            value = resolved.value
        elif type(resolved) == DynArray:
            _value = resolved.value
            value = []
            for i in _value:
                value.append(self.safely_resolve(i)) 
        elif type(resolved) == Map:
            _value = resolved.value
            value = {}
            for key in _value:
                value[
                    self.safely_resolve(key)
                ] = self.safely_resolve(_value[key])
                
        # Legacy
        # Markup safe breaks this a bit...
        elif type(resolved) in [int, str, list, tuple, float, markupsafe.Markup]:
            value = resolved
        elif type(resolved) == Auto:
            value = resolved.value
        else:
            print(f"Couldnt resolve {resolved} of {type(resolved)} in Module, returning None")

        print(value)

        return value
    
    def ref_resolve(self, var):
        # Dependencies
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')
        func_module: Module = self.compiler_instance.get_action('FUNCTION_CALL')
        if type(var) == tuple:
            resolved = resolution_module(var)
        else: resolved = var
        value: BasicType = None


        if type(resolved) == ID:
            value = resolved.value
            v = self.compiler_instance.get_variable(value)
            # print(v)
            if v["type"] == String:
                value = v['object']
            elif v["type"] is None:
                value = String("")
            else:
                value = v['object']
        elif type(resolved) == String:
            value = resolved
        elif type(resolved) == Int32:
            value = resolved
        elif type(resolved) == sInnerMut:
            value = func_module.run_sInnerMut(resolved)
        elif type(resolved) == SqlConnector:
            value = resolved
        elif type(resolved) == DynArray:
            value = resolved
        else:
            value = resolved
        return value


"""
# This module handles all functions related to Sessions.
# This module is based off of Flask sessions used in Spindle.
"""
from ..exceptions import TranspilerExceptions
from ..module import Module
from ..types import *

from flask import session
import random

SESSION_TRACKER = {}

def generate_session():
    global SESSION_TRACKER
    s = random.getrandbits(128)
    if s in SESSION_TRACKER:
        s = random.getrandbits(128)
    
    return "%032x" % s

class _SessionMod(Module):
    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def base(self, tree):
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')
        func_module: Module = self.compiler_instance.get_action('FUNCTION_CALL')
        values = []

        if 'POSITIONAL_ARGS' in tree['FUNCTION_ARGUMENTS']:
            for var in tree['FUNCTION_ARGUMENTS']['POSITIONAL_ARGS']:
                resolved = resolution_module(var)
                value: BasicType = None

                if type(resolved) == ID:
                    value = resolved.value
                    v = self.compiler_instance.get_variable(value)
                    if v["type"] is BasicType:
                        if v["type"] == String:
                            value = self.remove_quotes(v['object'].value)
                        else:
                            value = v['object'].value
                    elif v["type"] is None:
                        value = ""
                    else:
                        value = str(v['object'].value)
                elif type(resolved) == String:
                    value = self.remove_quotes(resolved.value)
                elif type(resolved) == Int32:
                    value = resolved.value
                elif type(resolved) == sInnerMut:
                    value = func_module.run_sInnerMut(resolved).value

                values.append(value)
        return values

class SessionStartMod(_SessionMod):
    name = "session_start"
    type = Module.MODULE_TYPES.FUNCTION

    def __init__(self, compiler_instance):
        super().__init__(compiler_instance)

    def proc_tree(self, tree):
        # TODO: Change prints to console logs
        values = self.base(tree)

        if len(values) != 0:
            raise TranspilerExceptions.IncorrectNumberOfValues(values, "session_start()")
        
        # session['aa'] = "bb"

        # Session already exists
        if "_noid" in session:
            print("Session exists")
        else:
            global SESSION_TRACKER
            try:
                session['_noid'] = generate_session()
            except RuntimeError as e:
                print("No secret key was set. Unable to securely create a session. Please add a 'secret_key' entry to the wool.yaml config.")
            SESSION_TRACKER[session['_noid']] = {
                "_cin": self.compiler_instance.namespace # Created in
            }
            print(f"Session created for {session['_noid']}")

        # Add new variable 
        self.compiler_instance.create_variable(
            "_SESSION",
            Session,
            Session(), 
            force=True
        )

        return Nil
    

class SessionDestroyMod(_SessionMod):
    name = "session_destroy"
    type = Module.MODULE_TYPES.FUNCTION

    def __init__(self, compiler_instance):
        super().__init__(compiler_instance)

    def proc_tree(self, tree):
        # TODO: Change prints to console logs
        values = self.base(tree)

        if len(values) != 0:
            raise TranspilerExceptions.IncorrectNumberOfValues(values, "session_destroy()")
        
        # session['aa'] = "bb"

        # Session already exists
        if "_noid" in session:
            print("Session exists, destroying")
            global SESSION_TRACKER
            if session['_noid'] in SESSION_TRACKER:
                del SESSION_TRACKER[session['_noid']]
            session.clear()
        return Nil


from cmath import inf
import os
import random
from flask import Flask, jsonify, send_file, send_from_directory
# Lang
from .lang.compiler import Compiler
from .lang.lexer import PyettyLexer
from .lang.pparser import PyettyParser
from .lang.types import Request
from .splitter import split_php

import json as _json

# Load modules
from .lang.modules import (
    BoolMod,
    ConditionalMod,
    ForEachMod,
    GetIndexMod,
    IncludeMod,
    InternalMod,
    MathMod,
    Module,
    NamespaceMemberMod,
    NamespaceMod,
    NewObjectMod,
    PhpMod,
    HTMLMod,
    ConcatMod,
    PrivateMod,
    ProtectedMod,
    PublicMod,
    RequireOnceMod,
    ReturnMod,
    SetIndexMod,
    TarrowMod,
    UseMod,
    VariableAssignMod,
    ResolutionMod,
    FunctionCallMod,
    ClassDeclarationMod,
    FunctionDecMod,
    WhileMod
)
# Language builtins
from .lang.std import (
    echo,
    rand,
    htmlspecialchars,
    string,
    session,
    redirect,
    primitives,
    db,
    bcrypt,
    json,
    internal,
)

# Read wool config
# from sys import argv
import argparse

import yaml
from yaml import Loader

parser = argparse.ArgumentParser(
    prog='NoPHP server',
    description=
'''NoPHP Spindle server''',
    epilog='Write Fast. Write More. Write NoPHP.')

parser.add_argument('name', default='')  
parser.add_argument('-o', '--host', default='localhost', help="Hostname to be used")
parser.add_argument('-p', '--port', default=8000, help="Port to listen on")
parser.add_argument('-d', '--debug', action='store_true', help="Debug mode, doesnt do much")
parser.add_argument('-n', '--project', action='store_true', help="Create a project. Requires a name")

args = parser.parse_args()
host = args.host
port = int(args.port)
debug = args.debug
conf = args.name

def create_example_project(project_root):
    # Create project root
    os.makedirs(project_root, exist_ok=True)

    # Create README file
    with open(os.path.join(project_root, "README"), "w") as readme_file:
        readme_file.write(f"Welcome to {project_root}!")

    # Create 'app' directory
    app_dir = os.path.join(project_root, "app")
    os.makedirs(app_dir, exist_ok=True)

    # Create 'viewcontrollers' directory inside 'app'
    viewcontrollers_dir = os.path.join(app_dir, "viewcontrollers")
    os.makedirs(viewcontrollers_dir, exist_ok=True)

    # Create 'index.php' file inside 'viewcontrollers'
    index_php_path = os.path.join(viewcontrollers_dir, "index.php")
    with open(index_php_path, "w") as index_php_file:
        index_php_file.write("<?php\n// Your NoPHP code goes here\n?>")

    # Create 'static' directory
    static_dir = os.path.join(project_root, "static")
    os.makedirs(static_dir, exist_ok=True)

    # Create 'config' directory
    config_dir = os.path.join(project_root, "config")
    os.makedirs(config_dir, exist_ok=True)

    # Create 'wool.yaml' file inside 'config'
    wool_yaml_path = os.path.join(config_dir, "wool.yaml")
    with open(wool_yaml_path, "w") as wool_yaml_file:
        wool_yaml_file.write(
f"""---
app: {project_root}
secret_key: CHANGEME_SECRET
routes:
  /: app/viewcontrollers/index.php
static:
  /files/<path:path>: static/"""
        )

    print(f"Project {project_root} was created!")

if args.project:
    create_example_project(conf)
    exit(0)

config = yaml.load(open(conf,"r"), Loader)

# Consts
BIGGEST_PAGES = 9_999_999_999

class SpindleApp:
    def __init__(self) -> None:
        self.app = Flask(__name__)
        self.lexer = PyettyLexer()
        self.parser = PyettyParser()

        self.functions = {}



    def build_sp(self, file, _c: Compiler):
        text = open(file,"r").read()
        php, html = split_php(text)
        toks = self.lexer.tokenize(
            php
        )
        ast = self.parser.parse(toks)



        _c.populate_modules_actions(
            [
                ConditionalMod,
                ForEachMod,
                GetIndexMod,
                InternalMod,
                MathMod,
                NamespaceMemberMod,
                NamespaceMod,
                NewObjectMod,
                PhpMod,
                HTMLMod,
                ConcatMod,
                PrivateMod,
                ProtectedMod,
                PublicMod,
                RequireOnceMod,
                ReturnMod,
                SetIndexMod,
                TarrowMod,
                UseMod,
                VariableAssignMod,
                ResolutionMod,
                FunctionCallMod,
                ClassDeclarationMod,
                FunctionDecMod,
                WhileMod,
                BoolMod,
            ]
        )
        _c.add_builtin_functions({
            # Fake internals as functions
            "echo": {
                "run_func": echo.EchoMod(_c)
            },
            "rand": {
                "run_func": rand.RandMod(_c)
            },
            "htmlspecialchars": {
                "run_func": htmlspecialchars.HTMLSpecialCharMod(_c)
            },
            "strlen": {
                "run_func": string.StrLenMod(_c)
            },
            "str_replace": {
                "run_func": string.StrReplaceMod(_c)
            },
            "substr": {
                "run_func": string.SubstrMod(_c)
            },
            "nl2br": {
                "run_func": string.Nl2BrMod(_c)
            },
            "session_start": {
                "run_func": session.SessionStartMod(_c)
            },
            "session_destroy": {
                "run_func": session.SessionDestroyMod(_c)
            },
            "redirect": {
                "run_func": redirect.RedirectMod(_c)
            },

            **db.build_funcs(_c),
            **primitives.build_funcs(_c),
            **bcrypt.build_funcs(_c),
            **json.build_funcs(_c),
            **internal.build_funcs(_c),
            
            "require_once": {
                "run_func": RequireOnceMod(_c)
            },
            "include": {
                "run_func": IncludeMod(_c)
            },
            "use": {
                "run_func": UseMod(_c)
            },
            "public": {
                "run_func": PublicMod(_c)
            },
            "private": {
                "run_func": PrivateMod(_c)
            },
            "protected": {
                "run_func": ProtectedMod(_c)
            }
        })
        c: Compiler = _c.new_instance(
            namespace='main',
            sync="b"
        )
        c.create_variable(
            "_SERVER",
            Request,
            Request()
        )
        c.run(ast)
        return '\n'.join([str(el) for el in c.finished]) + "\n" + html
        
    
    def register(self, route, file):
        global BIGGEST_PAGES
        def _func(*args, **kwargs):
            _c = Compiler([])
            out = self.build_sp(file, _c)
            return out
        
        name = f"_func_" + str(random.randint(0,BIGGEST_PAGES))
        while name in self.functions:
            name = f"_func_" + str(random.randint(0,BIGGEST_PAGES))


        _func.__name__ = _func.__qualname__ = name

        setattr(self, name, _func)
        print("Adding", _func.__name__, route)
        self.app.route(route, methods=['GET', 'POST'])(getattr(self, name))


    def register_static(self, route, file):
        global BIGGEST_PAGES
        def _func(path):
            print(file, path)
            return send_file(
                os.path.join(
                    os.getcwd(),
                    os.path.join(file, path)
                )
            )
        
        name = f"_sfunc_" + str(random.randint(0,BIGGEST_PAGES))
        while name in self.functions:
            name = f"_sfunc_" + str(random.randint(0,BIGGEST_PAGES))

        _func.__name__ = _func.__qualname__ = name

        setattr(self, name, _func)
        print("Adding", _func.__name__, route, file)
        self.app.route(route)(getattr(self, name))

    def register_json(self, route, file):
        global BIGGEST_PAGES
        def _func(*args, **kwargs):
            _c = Compiler([])
            out = self.build_sp(file, _c)
            try:
                return jsonify(_json.loads(out))
            except Exception:
                print(f"Failed to jsonify {out}")
        
        name = f"_jfunc_" + str(random.randint(0,BIGGEST_PAGES))
        while name in self.functions:
            name = f"_jfunc_" + str(random.randint(0,BIGGEST_PAGES))

        _func.__name__ = _func.__qualname__ = name

        setattr(self, name, _func)
        print("Adding", _func.__name__, route, file)
        self.app.route(route, methods=['GET', 'POST'])(getattr(self, name))
        
    def run(self):
        global host, port, debug
        print("Starting on {host}:{port}".format(host=host, port=port))
        self.app.run(host=host, port=port, debug=debug)
        
app = SpindleApp()
app.app.secret_key = config['secret_key'] if "secret_key" in config else ""
app.app.url_map.strict_slashes = False

for route in config['routes']:
    app.register(
        route,
        config['routes'][route]
    )
    
# if static routes need to be defined, enable them
if 'static' in config:
    for route in config['static']:
        app.register_static(
            route,
            config['static'][route]
        )

if 'json' in config:
    for route in config['json']:
        app.register_json(
            route,
            config['json'][route]
        )

if 'onstart' in config:
    _c = Compiler([])
    app.build_sp(config['onstart'], _c)

def main():
    app.run()

if __name__ == "__main__":
    main()


from copy import deepcopy
import inspect
import pprint
import re

from .exceptions import ModuleExceptions, TranspilerExceptions, RuntimeExceptions, Warn
from .module import Module
from nophp.splitter import split_php

from .lexer import PyettyLexer
from .pparser import PyettyParser
from .types import *
pprint = pprint.PrettyPrinter(indent=2).pprint
from rich.console import Console
from rich.syntax import Syntax
console = Console()

def rprint(data):
    console.print(
        Syntax(
            str(data),
            "Python", 
            padding=1, 
            line_numbers=True
        )
    )

class ConcatMod(Module):
    name="CONCAT"
    type = Module.MODULE_TYPES.ACTION

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')

        first, second = tree['0'], tree['1']

        supported_types = {
            ID,
            String,
            Int32,
            Bool,
            Auto,
            Nil
        }

        print(self.compiler_instance.namespace)

        first_resolved = resolution_module(first)
        second_resolved = resolution_module(second)


        # Todo: Legacy, remove as it's not used and may cause issues
        if type(first_resolved) == str: first_resolved = String(first_resolved)
        if type(second_resolved) == str: second_resolved = String(second_resolved)

        if type(first_resolved) not in supported_types or\
              type(second_resolved) not in supported_types:
            # Invalid type
            raise TranspilerExceptions.Generic(f"Not Implemented: {first_resolved} {second_resolved} Only supports: {supported_types}")
        
        first_value = ""
        if type(first_resolved) == ID:
            first_var: BasicType  = self.compiler_instance.get_variable(first_resolved.value)

            if first_var["type"] not in [String, Int32, Bool, Auto, Nil]:
                # Invalid type
                raise TranspilerExceptions.Generic(f"Expected an ID() of String() type, got {first_var['type'].__name__}()")
            
            first_value = first_var["object"].value
        elif type(first_resolved) == String:
            first_value = first_resolved.value
        elif type(first_resolved) ==  Int32:
            first_value = f"\"{first_resolved.value}\""
        elif type(first_resolved) == Auto:
            first_value = first_resolved.value


        second_value = ""
        if type(second_resolved) == ID:
            second_var: BasicType  = self.compiler_instance.get_variable(second_resolved.value)

            if second_var["type"] not in [String, Int32, Bool, Auto, Nil]:
                # Invalid type
                raise TranspilerExceptions.Generic(f"Expected an ID() of String() type, got {second_var['type'].__name__}()")
        
            second_value = second_var["object"].value
        elif type(second_resolved) == String:
            second_value = second_resolved.value
        elif type(second_resolved) ==  Int32:
            second_value = f"\"{second_resolved.value}\""
        elif type(second_resolved) == Auto:
            second_value = second_resolved.value

        return String(self.remove_quotes(first_value) + self.remove_quotes(second_value))

class PhpMod(Module):
    name="PHP"
    type = Module.MODULE_TYPES.NON_WRITEABLE

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        console.log("[PHP] triggered SP (single page) build.")

        # Match actions through local scope
        self.compiler_instance.run(
            tree['PROGRAM']
        )

class HTMLMod(Module):
    name="HTML"
    type = Module.MODULE_TYPES.SPECIAL_ACTION

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

        self.tag = ""
        self.attrs = {}

        from html.parser import HTMLParser
        from html.entities import name2codepoint

        class MyHTMLParser(HTMLParser):
            def __init__(self, host, convert_charrefs: bool = True) -> None:
                super().__init__(convert_charrefs=convert_charrefs)
                self.host = host
            def handle_starttag(self, tag, attrs):
                # print("Start tag:", tag)
                self.host.tag = tag
                for attr in attrs:
                    # print("     attr:", attr)
                    self.host.attrs[attr[0]] = attr[1]
        

        self.parser = MyHTMLParser(self)

    def proc_tree(self, tree):
        # pprint("[HTML] triggered simplified HTML build.")

        self.tag = ""
        self.attrs = {}

        if len(tree['PROGRAM']) == 0:
            value = ""
        elif tree['PROGRAM'][0] == 'HTML':
            value = self.__class__(self.compiler_instance)(tree['PROGRAM'][1])
        elif tree['PROGRAM'][0] == 'CONDITIONAL':
            instance = self.compiler_instance.new_instance(
                namespace= "_embedded_if",
                tree=(tree['PROGRAM'],),
                sync="advanced"
            )

            instance.run()
            value = '\n'.join([str(el) for el in instance.finished]) 
        else:
            obj = self.compiler_instance.get_action("RESOLUT")(tree['PROGRAM'])

            if type(obj) == String:
                # format string
                value = self.remove_quotes(obj.value)
            elif type(obj) == ID:
                value = self.remove_quotes(self.compiler_instance.get_variable(obj.value)['object'].value)
            elif obj is None:
                # Resolution returned None?
                raise TranspilerExceptions.Generic("Failed to parse")
            elif type(obj) == str: # TODO: Legacy
                value = obj
            else: value = obj.value


        # Check if it's a php call
        start = tree['LABEL']['START'] if tree['LABEL']['START'] is not None else ""
        end = tree['LABEL']['END'] if tree['LABEL']['END'] is not None else ""

        # Test parse of the start
        self.tag = ""
        self.attrs = {}
        self.parser.feed(start)
        # print(self.tag)
        # print(self.attrs)
        
        if self.attrs != {}:
            final = []
            # parse content of the attribute
            for attr in self.attrs:
                if self.attrs[attr] is None:
                    raise TranspilerExceptions.Generic(f"Invalid HTML at tag {self.tag}")
                if self.attrs[attr].startswith('$'):
                    code = f"echo {self.attrs[attr]}"
                    lexer = PyettyLexer()
                    parser = PyettyParser()
                    ast = parser.parse(lexer.tokenize(code))

                    instance = self.compiler_instance.new_instance(
                        tree=ast,
                        sync="advanced"
                    )
                    instance.run()
                    out = ''.join(instance.finished)
                    final.append(f"{attr}=\"{out}\"")
                else:
                    final.append(f"{attr}=\"{self.attrs[attr]}\"")
            start = f"<{self.tag} {' '.join(final)}>"

        # Todo: Legacy
        if type(value) == String:
            value = self.remove_quotes(value.value)

        return String(start + str(value) + end)
    
class NamespaceMod(Module):
    name="NAMESPACE"
    type = Module.MODULE_TYPES.SPECIAL_ACTION

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        # Dependencies
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')

        # print(tree)
        name = tree['ID']

        if name in self.compiler_instance.namespaces:
            # Continue using the namespace
            pass
        else:
            self.compiler_instance.namespace = name

            self.compiler_instance.namespaces[name] = self.compiler_instance


class NamespaceMemberMod(Module):
    name="NAMESPACE_MEMBER"
    type = Module.MODULE_TYPES.SPECIAL_ACTION

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        # Dependencies
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')

        # TODO: This will at some point need to loop over every namespace entry
        namespace = tree["NAMESPACE"]
        member = tree["MEMBER"]

        if namespace in self.compiler_instance.namespaces:
            ns = self.compiler_instance.namespaces[namespace]
        else:
            raise TranspilerExceptions.NamespaceNotFound(namespace, self.compiler_instance.namespaces)
        
        # TODO: Add Classes, other namespaces
        if member in ns.functions:
            return sInnerMut("namespace_member_func", member, ns.functions[member])
        if member in ns.classes:
            return sInnerMut("namespace_member_class", member, ns.classes[member])
        else:
            raise TranspilerExceptions.UnkownMethodReference(member, list(ns.functions.keys()))


class UseMod(Module):
    name="USE_FUNC"
    type = Module.MODULE_TYPES.NON_WRITEABLE

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def proc_tree(self, _):
        Warn("Use is obsolete in NoPHP, the namespace was autoloaded, skipping.")

class RequireOnceMod(Module):
    name="REQUIRE_ONCE"
    type = Module.MODULE_TYPES.NON_WRITEABLE

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        # Dependencies
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')
        values = []

        console.log("[red]<===>[/red] Require Once")

        if len(tree['FUNCTION_ARGUMENTS']['POSITIONAL_ARGS']) > 1:
            raise TranspilerExceptions.TooManyValues(values, "require_once($_path)")
        

        for var in tree['FUNCTION_ARGUMENTS']['POSITIONAL_ARGS']:
            resolved = resolution_module(var)
            value: BasicType = None
            
            if type(resolved) == ID:
                value = resolved.value
                v = self.compiler_instance.get_variable(value)
                if v["type"] == String:
                    value = self.remove_quotes(v['object'].value)
                elif v["type"] == type(None):
                    value = ""
                else:
                    value = v['object'].value
            elif type(resolved) == String:
                value = self.remove_quotes(resolved.value)
            

            values.append(value) 

        # Load file then lex and parse it
        tokens = None
        with open(value, "r") as f:
            text = f.read()
            php, html = split_php(text)

            lexer = PyettyLexer()
            parser = PyettyParser()

            tokens = parser.parse(
                lexer.tokenize(php)
            )


        old_nm = self.compiler_instance.namespace

        instance = self.compiler_instance.new_instance(
            namespace= value+"_sub",# Set namespace
            tree=tokens, # Load file here with lexer & parser
            sync="advanced"
        )

        instance.run()

        self.compiler_instance.namespace = old_nm

        self.compiler_instance.namespaces[
            instance.namespace
        ] = instance

        self.compiler_instance.write_back(
            instance
        )
        # self.compiler_instance.sync_variables(
        #     instance
        # )

        # Sync functions to own instance
        self.compiler_instance.sync_functions(
            instance
        )


class IncludeMod(Module):
    name="INCLUDE"
    type = Module.MODULE_TYPES.SPECIAL_ACTION

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        # Dependencies
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')
        values = []

        console.log("[red]<===>[/red] Include")


        Warn("Include is broken, please make sure to use the latest RC versions for fastest patches")

        if len(tree['FUNCTION_ARGUMENTS']['POSITIONAL_ARGS']) > 1:
            raise TranspilerExceptions.TooManyValues(values, "include($_path)")
        

        for var in tree['FUNCTION_ARGUMENTS']['POSITIONAL_ARGS']:
            resolved = resolution_module(var)
            value: BasicType = None
            
            if type(resolved) == ID:
                value = resolved.value
                v = self.compiler_instance.get_variable(value)
                if v["type"] == String:
                    value = self.remove_quotes(v['object'].value)
                elif v["type"] == type(None):
                    value = ""
                else:
                    value = v['object'].value
            elif type(resolved) == String:
                value = self.remove_quotes(resolved.value)
            

            values.append(value) 

        # Load file then lex and parse it
        tokens = None
        with open(value, "r") as f:
            text = f.read()
            php, html = split_php(text)

            lexer = PyettyLexer()
            parser = PyettyParser()

            tokens = parser.parse(
                lexer.tokenize(php)
            )


        old_nm = self.compiler_instance.namespace

        instance = self.compiler_instance.new_instance(
            namespace= value+"_sub",# Set namespace
            tree=tokens, # Load file here with lexer & parser
            sync=""
        )

        instance.finished = []

        instance.run()

        self.compiler_instance.namespace = old_nm

        self.compiler_instance.namespaces[
            instance.namespace
        ] = instance


        return String('\n'.join([str(el) for el in self.compiler_instance.finished]))
        

class GetIndexMod(Module):
    name="GET_INDEX"
    type = Module.MODULE_TYPES.SPECIAL_ACTION

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        # Dependencies
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')

        # pprint(tree)
        line = tree['LINE']
        name = resolution_module(tree['EXPRESSION'])
        index = resolution_module(tree['INDEX'])

        # print(index)
        # print(name)

        if type(index) == String:
            index_value = self.remove_quotes(index.value)
        else: index_value = index.value

        if type(name) == ID:
            var = self.compiler_instance.get_variable(name.value)
            # print(var)
            if var['type'] == DynArray:
                if 0 <= index_value < var['object'].length:
                    return var['object'].value[index_value]
                else:
                    print(var['object'])
                    raise TranspilerExceptions.OutOfBounds(index_value, var['object'].length)
            # DONE: Session and Request types need to be compatible with printing out in echo
            elif var['type'] == Session:
                if index_value in var['object'].value:
                    # print("Session contains:",var['object'].value[index_value], type(var['object'].value[index_value]), var['object'].value)
                    return Auto(var['object'].value[index_value], "basic")
                else:
                    # print(f"Session does not contain {index_value}, it only contains: {list(var['object'].value.keys())}")
                    return Nil()
            elif var['type'] == Request:
                if index_value.lower() in var['object'].value.__dir__():
                    # print("Request contains:", getattr(var['object'].value, index_value.lower()), index_value.lower(), list(var['object'].value.__dir__()))
                    return Auto(getattr(var['object'].value, index_value.lower()), type="basic")
                else:
                    # print("Request contains:",list(var['object'].value.__dir__()))
                    return Nil()
            elif var['type'] == Map:
                if index_value in var['object'].value:
                    print("Found mapping")
                    return Auto(var['object'].value[index_value], "basic")
                else:
                    return Nil()
            else:
                raise TranspilerExceptions.TypeMissmatch("Get ID Index Actor", var['type'], [ID, DynArray, Session, Request], line)
        elif type(name) == DynArray:
            if 0 <= index_value < name.length:
                return Auto(name.value[index_value], "basic")
            else:
                raise TranspilerExceptions.OutOfBounds(index_value, name.length)
            # raise TranspilerExceptions.Generic(f"{name.value}")
        elif type(name) == Auto:
            # This is dumb but assume we got passed some internal object right?
            from werkzeug.datastructures import ImmutableMultiDict
            if type(name.value) == ImmutableMultiDict:
                if index_value in name.value.to_dict():
                    return Auto(name.value.to_dict()[index_value], "basic")
                else:
                    # Not present
                    return Nil()
            elif type(name.value) == dict:
                if index_value in name.value:
                    return Auto(name.value[index_value], "basic")
                else:
                    return Nil()
            elif type(name.value) == Map:
                if index_value in name.value.value:
                    return Auto(name.value.value[index_value], "basic")
                else:
                    return Nil()
            elif type(name.value) == DynArray:
                if 0 <= index_value < name.value.length:
                    return Auto(name.value.value[index_value], "basic")
                else:
                    raise TranspilerExceptions.OutOfBounds(index_value, name.value.length)
            # raise TranspilerExceptions.Generic(f"Can't get index on {name.value}")
            return Nil()
        elif type(name) == Nil:
            return Nil()
        else:
            raise TranspilerExceptions.TypeMissmatch(f"Get Index Actor: {name}", type(name), [ID, DynArray], line)
        return Nil()

class SetIndexMod(Module):
    name="SET_INDEX"
    type = Module.MODULE_TYPES.SPECIAL_ACTION

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        # Dependencies
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')

        # pprint(tree)
        # exit()
        line = tree['ID'][-1]
        name = resolution_module(tree['ID'][1]['EXPRESSION'])
        index = resolution_module(tree['ID'][1]['INDEX'])
        value = resolution_module(tree['EXPRESSION'])

        if type(index) == String:
            index_value = self.remove_quotes(index.value)
        else: index_value = index.value

        if type(value) == String:
            value = self.remove_quotes(value.value)
        

        if type(name) == ID:
            var = self.compiler_instance.get_variable(name.value)
            if var['type'] == DynArray:
                if 0 <= index_value < var['object'].length:
                    var['object'].value[index_value] = value
                else:
                    raise TranspilerExceptions.OutOfBounds(index_value, var['object'].length)
            elif var['type'] == Session:
                # print("Session contains:",var['object'].value[index_value], type(var['object'].value[index_value]))
                var['object'].value[index_value] = value.value
            else:
                raise TranspilerExceptions.TypeMissmatch("Get ID Index Actor", var['type'], [ID, DynArray], line)
        elif type(name) == DynArray:
            raise TranspilerExceptions.NotImplemented
        elif type(name) == Auto:
            if type(name.value) == dict:
                name.value[index_value] = value
            elif type(name.value) == Map:
                name.value.value[index_value] = value
            else:
                raise TranspilerExceptions.Generic(f"{name.value} is not a map.")
        else:
            raise TranspilerExceptions.TypeMissmatch(f"Get Index Actor '{name}'", type(name), [ID, DynArray], line)
        
class ResolutionMod(Module):
    name="RESOLUT"
    type = Module.MODULE_TYPES.ACTION

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance
    
    def proc_tree(self, tree):
        #pprint(tree)
        
        if tree[0] == 'STRING':
            return String(tree[1]['VALUE'])

        if tree[0] == 'FORMATTED_STRING':
            pattern = re.compile(r":\$(\w+):")
            text = tree[1]['VALUE']

            def replace(match):
                variable_name = match.group(1)
                value = self.compiler_instance.get_variable(variable_name)['object']
                
                return str(self.safely_resolve(value.value))

            result = pattern.sub(replace, text)
            return String(result)

        elif tree[0] == 'INT':
            return Int32(tree[1]['VALUE'])
        
        elif tree[0] == 'NEG':
            return Int32(-(self(tree[1]).value))

        elif tree[0] == 'FLOAT':
            return Float(tree[1]['VALUE'])
        
        elif tree[0] == 'BOOL':
            return Bool(tree[1]['VALUE'])

        elif tree[0] == 'HEX':
            return HexInt32(tree[1]['VALUE'])
        
        elif tree[0] == 'Nil':
            return Nil(0)
        
        elif tree[0] == 'NULL':
            return Nil(0)

        elif tree[0] == 'CHAR':
            # if the actual type is string, there will be an array in place of the value
            if type(tree[1]['VALUE']) == tuple:
                return Char(tree[1]['VALUE'][0])
            else:
                return Char(tree[1]['VALUE'])
            
        elif tree[0] == 'ARRAY':
            return DynArray(tree[1]['ITEMS'], type=List.BONE)
        
        elif tree[0] == 'ARRAY_FILLED':
            items = [self.proc_tree(item) for item in tree[1]['ITEMS']]
            return DynArray(items, type=List.GUESS)
        
        elif tree[0] == 'ASSOC_ARRAY':
            items = {}
            for item in tree[1]['ITEMS']:
                key = self.proc_tree(item[0])
                value = self.proc_tree(item[1])
                items[key] = value
            return Map(items)
        
        elif tree[0] == 'MAP':
            if len(tree[1]['ITEMS']) == 0:
                items = dict()
            else:
                raise TranspilerExceptions.NotImplemented
            return Map(items)

        elif tree[0] == 'ID':
            return ID(tree[1]['VALUE'])
        
        elif tree[0] == 'CONCAT':
            concat_module: Module = self.compiler_instance.get_action('CONCAT')
            return concat_module(tree[1])
        
        elif tree[0] == 'HTML':
            html_module: Module = self.compiler_instance.get_action("HTML")
            return html_module(tree[1])
        
        elif tree[0] == 'FUNCTION_CALL':
            stack = inspect.stack()

            # print("Resolution on", tree[1]['ID'], "by", f"{stack[2].filename}:{stack[2].lineno}")
            func_call_mod: Module = self.compiler_instance.get_action("FUNCTION_CALL")
            return func_call_mod(tree[1])
        
        elif tree[0] == 'NEW_OBJECT':
            new_object_mod: Module = self.compiler_instance.get_action("NEW_OBJECT")
            return new_object_mod(tree[1])
        
        elif tree[0] == 'GET_INDEX':
            get_index_mod: Module = self.compiler_instance.get_action("GET_INDEX")
            return get_index_mod(tree[1])
        
        elif tree[0] in {
            "ADD", "SUB", "MUL", "DIV"
        }:
            math_mod: Module = self.compiler_instance.get_action("MATH")
            return math_mod(tree)
        
        elif tree[0] == "NAMESPACE_MEMBER":
            NamespaceMemberMod: Module = self.compiler_instance.get_action("NAMESPACE_MEMBER")
            return NamespaceMemberMod(tree[1])
        
        elif tree[0] == "TARROW":
            TarrowMod: Module = self.compiler_instance.get_action("TARROW")
            return TarrowMod(tree[1])
        
        elif tree[0] == "FOREACH":
            ForeachMod: Module = self.compiler_instance.get_action("FOREACH")
            return ForeachMod(tree[1], no_construct=True)
        
        elif tree[0] in ["OR", "AND"]:
            BoolMod: Module = self.compiler_instance.get_action("BOOLMOD")
            return BoolMod(tree)
        

        elif tree[0] in [
            'EQEQ',
            'NOT_EQEQ',
            'LESS',
            'GREATER'
            ]:
            CondMod: Module = self.compiler_instance.get_action("CONDITIONAL")
            return CondMod(tree)
        
        elif tree[0] == "CONDITIONAL":
            CondMod: Module = self.compiler_instance.get_action("CONDITIONAL")
            return CondMod(tree[1])
        
        raise Exception(f"[RESOLUT] Failed to match {tree} is '{tree[0]}' supported?")


class ClassDeclarationMod(Module):
    name="CLASS_DECLARATION"
    type = Module.MODULE_TYPES.ACTION

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        # pprint(tree)
        extends = None
        source: list = []

        if "EXTENDS" in tree:
            # Check if the class exists
            extends = tree["EXTENDS"]

            if extends in self.compiler_instance.classes:
                # Get tree of the class, then we copy its code
                instance = self.compiler_instance.new_instance(
                    namespace=tree['ID']
                )
                source = list(self.compiler_instance.classes[extends]["source"])
                source.extend(tree['PROGRAM'])
                instance.run(source)
                extends = self.compiler_instance.classes[extends]['object']
            else:
                raise TranspilerExceptions.ClassNotFound(extends)

        else:

            # Create a new instance, it will track the functions and variables in its namespace
            instance = self.compiler_instance.new_instance(
                namespace=tree['ID'],
                tree=tree['PROGRAM']
            )
            source = tree['PROGRAM']
            instance.run()

        # Then Add this instance as a class template object, we can duplicate it later
        self.compiler_instance.create_class(
            name=tree['ID'],
            obj=instance,
            source=source,
            parent=extends
        )

        # print(
        #     instance.variables.keys()
        # )
        # input('c')

        console.log("[ClassMod] New class & scope created for \"{}\" ({})".format(tree['ID'], instance))

class NewObjectMod(Module):
    name="NEW_OBJECT"
    type = Module.MODULE_TYPES.ACTION

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        # pprint(tree)
        # Dependencies
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')

        name = resolution_module(tree['ID'])
        arguments = tree['FUNCTION_ARGUMENTS']

        if type(name) == sInnerMut:
            clas = name.value[2]
            obj = self.compiler_instance.new_instance(
                    namespace=name.value[1] + "_obj",
                    sync="b"
                )
            obj.run(clas['source'])
        elif name.value in self.compiler_instance.classes:
            obj = self.compiler_instance.new_instance(
                    namespace=name.value + "_obj",
                    sync="b"
                )
            obj.run(self.compiler_instance.classes[name.value]["source"])
        else:
            raise TranspilerExceptions.ClassNotFound(name)

        # Call it's constructor
        if "__construct" in obj.functions:
            if 'POSITIONAL_ARGS' in arguments:
                parsed_args = [
                    resolution_module(arg)
                    for arg in arguments['POSITIONAL_ARGS']
                ]

                for i, arg in enumerate(obj.functions["__construct"]["arguments"]):
                    try:
                        v = parsed_args[i]
                    except IndexError as e:
                        print(parsed_args)
                        # raise TranspilerExceptions.OutOfBounds(i, len(parsed_args))
                        raise TranspilerExceptions.Generic(f"Too few arguments provided for constructor.\n\tGot:{parsed_args}\n\tExpected:{obj.functions["__construct"]["arguments"]}")

                    if type(v) == ID:
                        n = v.value
                        v = self.compiler_instance.get_variable(v.value)['object']
                    else: n = None

                    # print(n, v, self.compiler_instance.namespace)


                    obj.functions["__construct"]["object"].create_variable(
                        arg.value, 
                        type(v),
                        v,
                        force=True
                    )

            if obj.functions["__construct"]['level'] != "public" \
                and self.compiler_instance.namespace != obj.functions["__construct"]['object'].parent.namespace:
                raise TranspilerExceptions.Generic(f"Invalid access level on constructor of {obj.namespace}")

            func = obj.functions["__construct"]["run_func"]
            func()

            # pprint(obj.functions["__construct"]['object'].variables)

            self.compiler_instance.write_back(
                obj.functions["__construct"]['object']
            )
            # print(obj.variables)
            # print(obj.functions["__construct"]['object'].variables)
            # console.log("Ran constructor for \"{}\"\n{}".format(name, 
            #             self.compiler_instance.classes[name]["source"]))

        console.log("New object created for \"{}\"(...)".format(name))

        return obj



class FunctionDecMod(Module):
    name="FUNCTION_DECLARATION"
    type = Module.MODULE_TYPES.SPECIAL_ACTION_FUNC

    def __init__(self, compiler_instance):
        super().__init__()

        self.compiler_instance = compiler_instance
    
    def proc_tree(self, tree):
        # Dependencies
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')

        line = tree['ID'][-1]

        funcname = tree['ID']
        arguments = tree['FUNCTION_ARGUMENTS']
        program  = tree['PROGRAM']
        returns  = tree['RETURNS_TYPE'][1]['VALUE'] if "RETURNS_TYPE" in tree else "any"

        types = {
            'int': Int32,
            'hex': HexInt32,
            'char': Char,
            'float': Float,
            'string': String,
            'void': Nil,
            'any': Any
        }

        # pprint(program)
        # exit()

        if arguments:
            arguments = [resolution_module(arg) for arg in arguments['POSITIONAL_ARGS']]
        
        console.log(f"[FunctionDecMod] {funcname}({arguments}) -> {returns} in {self.compiler_instance.namespace}")

        # Create a new instance, it will be filled with the function's tree
        instance = self.compiler_instance.new_instance(
            namespace=self.compiler_instance.namespace+"_"+funcname+"_ns",
            tree=program,
            sync = "object"
        )

        instance.parent = self.compiler_instance

        if 'LEVEL' in tree:
            if self.compiler_instance.namespace != 'main':
                level = tree['LEVEL']
            else:
                raise TranspilerExceptions.IncompleteDeclaration(f"Access levels are only allowed in classes, not {self.compiler_instance.namespace}")
        else: level = "public"

        # On call run the instance 
        # with injected locals for arguments
        #
        # instance.run()

        console.log(f"{funcname}(...) is built in '{self.compiler_instance.namespace}'")

        self.compiler_instance.create_function(
            funcname,
            instance,
            tree,
            types[returns],
            arguments=arguments,
            level = level 
        )


class FunctionCallMod(Module):
    name="FUNCTION_CALL"
    type = Module.MODULE_TYPES.SPECIAL_ACTION

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def run_sInnerMut(self, funcobj, arguments):
        # Dependencies
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')
        if 'POSITIONAL_ARGS' in arguments:
            parsed_args = [
                resolution_module(arg)
                for arg in arguments['POSITIONAL_ARGS']
            ]
        else: parsed_args = []
        if funcobj.value[0] == "namespace_member_func":
            for i, arg in enumerate(funcobj.value[2]["arguments"]):
                try:
                    v = parsed_args[i]
                except IndexError as e:
                    raise TranspilerExceptions.OutOfBounds(f"{i}, {arg}, {parsed_args}", len(funcobj.value[2]["arguments"]))
        
                funcobj.value[2]['object'].create_variable(
                    arg.value, 
                    type(v),
                    v,
                    force=True
                )
                # print(v.value)
            # We are executing something from a different namespace
            # ("namespace_member_func", member, ns.functions[member])
            func = funcobj.value[2]["run_func"]
            # funcobj.value[2]['object'].stop = False
            func()
            self.compiler_instance.write_back(
                funcobj.value[2]['object']
            )
            ret = funcobj.value[2]['object'].returns
            exp = funcobj.value[2]['returns']

            if ret != Nil:
                if type(ret) == ID:
                    ret = funcobj.value[2]['object'].get_variable(ret.value)['object']
                if type(ret) != exp and exp != Any:
                    raise TranspilerExceptions.TypeMissmatchNL(f"{funcobj.value[1]}(...) return value", ret, exp)
                # print(ret.value)
                # input()
                return ret
        else:
            raise TranspilerExceptions.NotImplemented()

    def proc_tree(self, tree):
        # TODO: Simplify this
        #           - Move all expected return validation into a separate function
        #           - Parse args here instead of inside the function
        #           - Raise proper error when a parent doesnt exist
        

        # Dependencies
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')
        done = None
        
        supported_types = {
            String,
            Nil,
            Int32,
            Float,
            Char,
            List,
            ID
        }

        if 'CLASS' in tree:
            # We assume it's part of a class
            #    {"FUNCTION_ARGUMENTS": p.function_arguments, "ID": p.ID1, "CLASS": p.ID0},
            cls = tree['CLASS']
            arguments = tree['FUNCTION_ARGUMENTS']
            funcname = tree['ID']
            funcobj = None
        else:
            # pprint(tree['ID'])
            # input()
            funcobj = resolution_module(tree['ID'])
            funcname = funcobj.value
            arguments = tree['FUNCTION_ARGUMENTS']
            cls = None

        if type(funcobj) != sInnerMut and funcname not in self.compiler_instance.builtin_functions:
            if 'POSITIONAL_ARGS' in arguments:
                parsed_args = [
                    resolution_module(arg)
                    for arg in arguments['POSITIONAL_ARGS']
                ]
            else: parsed_args = []

        console.log(f"[FunctionCallMod] {funcname}(...) from {cls}")

        if type(funcobj) == sInnerMut:
            return self.run_sInnerMut(funcobj, arguments)
        elif cls is not None:
            if cls == 'parent':
                function_parent = self.compiler_instance.parent
                class_parent    = function_parent.parent # grandad
                if class_parent is None:
                    raise TranspilerExceptions.ClassNotFound(f"None. This class does not contain a parent: {function_parent.namespace}")
                if funcname in class_parent.functions:
                    class_parent.functions[funcname]['run_func']()
                    # Sync back

                    self.compiler_instance.write_back(
                        class_parent.functions[funcname]['object']
                    )
                    ret = class_parent.functions[funcname]['object'].returns
                    exp = class_parent.functions[funcname]["returns"]
                    if ret != Nil:
                        if type(ret) != exp and exp != Any:
                            raise TranspilerExceptions.TypeMissmatch(f"{funcname}(...) return value", ret, exp, tree['ID'][-1])
                        done = ret
                else:
                    raise TranspilerExceptions.UnkownMethodReference(funcname, class_parent.functions)
                
            elif cls == 'this':
                function_parent = self.compiler_instance.parent

                if function_parent is None:
                    raise TranspilerExceptions.ClassNotFound(cls)
                
                if funcname in function_parent.functions:
                    function_parent.functions[funcname]['run_func']()
                    # Sync back

                    self.compiler_instance.write_back(
                        function_parent.functions[funcname]['object']
                    )
                    ret = function_parent.functions[funcname]['object'].returns
                    exp = function_parent.functions[funcname]["returns"]
                    if ret != Nil:
                        if type(ret) != exp and exp != Any:
                            raise TranspilerExceptions.TypeMissmatch(f"{funcname}(...) return value", ret, exp, tree['ID'][-1])
                        done = ret
                else:
                    raise TranspilerExceptions.UnkownMethodReference(funcname, function_parent.functions)

            elif cls in self.compiler_instance.variables:
                if funcname in self.compiler_instance.variables[cls]['object'].functions:
                    funcobj = self.compiler_instance.variables[cls]['object'].functions[funcname]

                    for i, arg in enumerate(self.compiler_instance.variables[cls]['object'].functions[funcname]["arguments"]):
                        try:
                            v = parsed_args[i]
                        except IndexError as e:
                            raise TranspilerExceptions.OutOfBounds(i, len(self.compiler_instance.variables[cls]['object'].functions[funcname]["arguments"]))

                        self.compiler_instance.variables[cls]['object'].functions[funcname]["object"].create_variable(
                            arg.value, 
                            type(v),
                            v,
                            force=True
                        )
                    # TODO: REMOVE
                    # print(self.compiler_instance.variables[cls]['object'].functions[funcname]["object"].namespace ,self.compiler_instance.variables[cls]['object'].functions[funcname]["object"].variables)
                    func = self.compiler_instance.variables[cls]['object'].functions[funcname]["run_func"]
                    # TODO: Check permission level
                    func()
                    self.compiler_instance.write_back(
                        self.compiler_instance.variables[cls]['object'].functions[funcname]['object']
                    )
                    ret = self.compiler_instance.variables[cls]['object'].functions[funcname]['object'].returns
                    exp = self.compiler_instance.variables[cls]['object'].functions[funcname]['returns']
                    # TODO: REMOVE
                    # print(funcname, ret)
                    if ret != Nil:
                        if type(ret) == ID:
                            # print(list(self.compiler_instance.variables[cls]['object'].functions[funcname]['object'].variables.keys()))
                            # input("vars.")
                            ret = self.compiler_instance.variables[cls]['object'].functions[funcname]['object'].get_variable(ret.value)['object']
                        if type(ret) != exp and exp != Any:
                            raise TranspilerExceptions.TypeMissmatchNL(f"{funcname}(...) return value", ret, exp)
                        done = ret
                else:
                    raise TranspilerExceptions.UnkownMethodReference(funcname, list(self.compiler_instance.variables[cls]['object'].functions.keys()), ns=self.compiler_instance.variables[cls]['object'].namespace)
            else:
                raise TranspilerExceptions.ClassNotFound(cls)
            
        elif funcname in self.compiler_instance.builtin_functions:
            done = self.compiler_instance.builtin_functions[funcname]["run_func"](tree)

        elif funcname in self.compiler_instance.functions:
            for i, arg in enumerate(self.compiler_instance.functions[funcname]["arguments"]):
                try:
                    v = parsed_args[i]
                except IndexError as e:
                    raise TranspilerExceptions.OutOfBounds(i, len(self.compiler_instance.functions[funcname]["arguments"]))

                self.compiler_instance.functions[funcname]["object"].create_variable(
                    arg.value, 
                    type(v),
                    v,
                    force=True
                )

            self.compiler_instance.functions[funcname]["run_func"]()
            # Sync back
            self.compiler_instance.write_back(
                self.compiler_instance.functions[funcname]["object"]
            )
            ret = self.compiler_instance.functions[funcname]["object"].returns
            exp = self.compiler_instance.functions[funcname]["returns"]
            if ret != Nil:
                if type(ret) != exp and exp != Any:
                    raise TranspilerExceptions.TypeMissmatch(f"{funcname}(...) return value", ret, exp, tree['ID'][-1])
                done = ret
        else:
            raise TranspilerExceptions.UnkownMethodReference(funcname, list(self.compiler_instance.functions.keys()), self.compiler_instance.namespace)
        
        return done


class ReturnMod(Module):
    name="RETURN"
    type = Module.MODULE_TYPES.ACTION

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance
    
    def proc_tree(self, tree):
        # Dependencies
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')

        v = resolution_module(tree["EXPRESSION"])

        self.compiler_instance.returns = v
        self.compiler_instance.stop = True
        


class TarrowMod(Module):
    name="TARROW"
    type = Module.MODULE_TYPES.NON_WRITEABLE

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance
    
    def proc_tree(self, tree):
        # TODO: Add sometihng idk
        # TODO: Add parent
        if tree["FROM"] == "this":
            # print(self.compiler_instance.get_variable(tree["TO"])['object'])
            return self.compiler_instance.get_variable(tree["TO"])['object']
        elif tree["FROM"] in self.compiler_instance.variables:
            o = self.compiler_instance.get_variable(tree["FROM"])['object']
            # print(tree["TO"], o.get_variable(tree["TO"]), self.compiler_instance.namespace)
            return Auto(o.get_variable(tree["TO"])['object'], 'basic').value
        else:
            raise TranspilerExceptions.NotImplemented
        
class ClassAttrMod(Module):
    name="CLASS_ATTR"
    type = Module.MODULE_TYPES.NON_WRITEABLE

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        # Dependencies
        raise TranspilerExceptions.NotImplemented
    

### ACCESS LEVELS ###
# TODO: ACTUALLY IMPLEMENT THEM
class PublicMod(Module):
    name="public"
    type = Module.MODULE_TYPES.NON_WRITEABLE

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        # Dependencies
        varas_module: Module = self.compiler_instance.get_action('VARIABLE_ASSIGNMENT')

        # Create var :)))
        ntree = {}
        ntree['EXPRESSION'] = ["Nil"]
        ntree['ID'] = tree['FUNCTION_ARGUMENTS']['POSITIONAL_ARGS'][0][1]['VALUE']

        varas_module(ntree)

class ProtectedMod(Module):
    name="protected"
    type = Module.MODULE_TYPES.NON_WRITEABLE

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        # Dependencies
        varas_module: Module = self.compiler_instance.get_action('VARIABLE_ASSIGNMENT')

        # Create var :)))
        ntree = {}
        ntree['EXPRESSION'] = ["Nil"]
        ntree['ID'] = tree['FUNCTION_ARGUMENTS']['POSITIONAL_ARGS'][0][1]['VALUE']

        varas_module(ntree)


class PrivateMod(Module):
    name="private"
    type = Module.MODULE_TYPES.NON_WRITEABLE

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        # Dependencies
        varas_module: Module = self.compiler_instance.get_action('VARIABLE_ASSIGNMENT')

        # Create var :)))
        ntree = {}
        ntree['EXPRESSION'] = ["Nil"]
        ntree['ID'] = tree['FUNCTION_ARGUMENTS']['POSITIONAL_ARGS'][0][1]['VALUE']

        varas_module(ntree)

###############################################################################################


class VariableAssignMod(Module):
    name="VARIABLE_ASSIGNMENT"
    type = Module.MODULE_TYPES.NON_WRITEABLE

    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def proc_tree(self, tree):
        # Dependencies
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')

        expr = tree['EXPRESSION']
        var  = tree['ID']
        parent = tree['PARENT'] if 'PARENT' in tree else ""
        # We will need to guess the type from the expression we are given
        # vartype = tree['TYPE'] 


        resolved = resolution_module(expr)

        console.log(f"[VariableAssignMod] '{var}': guess -> {expr} to {resolved} in {self.compiler_instance.namespace}")

        while type(resolved) == ID:
            resolved = self.compiler_instance.get_variable(resolved.value)['object']

        if var not in self.compiler_instance.variables:
            # We create a new variable
            self.compiler_instance.create_variable(var, type(resolved), resolved)
        else:
            self.compiler_instance.create_variable(var, type(resolved), resolved, force=True)
        
        # No need to return anything here
        # The module is of type Non-Writeable



class MathMod(Module):
    name="MATH"
    type = Module.MODULE_TYPES.ACTION

    def __init__(self, compiler_instance):
        super().__init__()

        self.compiler_instance = compiler_instance

    def add(self, val1, val2):
        return val1.value + val2.value

    def sub(self, val1, val2):
        return val1.value - val2.value
    
    def mult(self, val1, val2): 
        return val1.value * val2.value

    def div(self, val1, val2): 
        return val1.value / val2.value
    
    def proc_tree(self, tree):
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')

        ops = {
            'ADD': self.add,
            'SUB': self.sub,
            'MUL': self.mult,
            'DIV': self.div
        }

        op = ops[tree[0]]

        # Process first and second value
        first = resolution_module(tree[1])
        second = resolution_module(tree[2])
        
        def atomize(val):
            if type(val) == ID:
                val = self.compiler_instance.get_variable(val.value)['object']
            return val
        
        first = atomize(first)
        second = atomize(second)

        return Int32(op(first, second))

class ConditionalMod(Module):
    name="CONDITIONAL"
    type = Module.MODULE_TYPES.SPECIAL_ACTION

    def __init__(self, compiler_instance):
        super().__init__()

        self.compiler_instance = compiler_instance

    def equal(self, val1, val2):
        return self.remove_quotes(val1) == self.remove_quotes(val2)

    def nequal(self, val1, val2):
        return not self.equal(val1, val2)
    
    def less(self, val1, val2): 
        return self.remove_quotes(val1) < self.remove_quotes(val2)

    def greater(self, val1, val2): 
        return self.remove_quotes(val1) > self.remove_quotes(val2)
    
    def function_call(self, tree):
        func_module: Module = self.compiler_instance.get_action('FUNCTION_CALL')
        res = func_module(tree)
        return res.value
    
    def or_adv_bool(self, val1, val2):
        return val1 or val2
    
    def and_adv_bool(self, val1, val2):
        return val1 and val2
    
    def proc_tree(self, tree):
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')

        if 'IF' in tree:
            if_condition = tree['IF'][1]['CONDITION']
            if_program = tree['IF'][1]['CODE']
            else_pr = None if tree["ELSE"][0] is None else tree["ELSE"]
        else:
            if_condition = tree
            if_program = None
            else_pr = None

        conditionals = {
            'EQEQ': self.equal,
            'NOT_EQEQ': self.nequal,
            'LESS': self.less,
            'GREATER': self.greater,
            'FUNCTION_CALL': self.function_call,
            'OR': self.or_adv_bool,
            'AND': self.and_adv_bool
        }


        # Resolve condition
        condition = conditionals[if_condition[0]]


        if if_condition[0] != 'FUNCTION_CALL':
            # Process first and second value
            first = resolution_module(if_condition[1])
            second = resolution_module(if_condition[2])

            def atomize(val):
                if type(val) == ID:
                    val = atomize(self.compiler_instance.get_variable(val.value)['object'])
                elif type(val) in [Int32, String, Bool, Float] :
                    val = val.value
                elif type(val) == Auto:
                    val = atomize(val.value)
                else:
                    print(type(val))
                return val

            first = atomize(first)
            second = atomize(second)

            # Evaluate condition and run it's code
            c = condition(first, second)
            # print(c, self.remove_quotes(first), self.remove_quotes(second), if_condition[0])
        else:
            c = condition(if_condition[1])
            # print(c)
        if c and if_program is not None:
            self.compiler_instance.run(if_program)
            return Bool(True)
        else:
            if else_pr is not None:
                self.compiler_instance.run(else_pr[1]['CODE'])
            return Bool(False)
        
        
class BoolMod(Module):
    name="BOOLMOD"
    type = Module.MODULE_TYPES.SPECIAL_ACTION

    def __init__(self, compiler_instance):
        super().__init__()

        self.compiler_instance = compiler_instance
    
    def proc_tree(self, tree):
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')
        conditional_module: Module = self.compiler_instance.get_action('CONDITIONAL')
        empty: Module = self.compiler_instance.get_action('empty') # Actually a function

        pprint(tree)
        op = tree[0]
        first = tree[1]
        second = tree[2]

        if op == 'AND':
            return empty((first,)).value and empty((second,)).value
        elif op == 'OR':
            return empty((first,)).value or empty((second,)).value
        else:
            raise TranspilerExceptions.NotImplemented

        quit()


class WhileMod(Module):
    name="WHILE"
    type = Module.MODULE_TYPES.ACTION

    def __init__(self, compiler_instance):
        super().__init__()

        self.compiler_instance = compiler_instance

    
    def proc_tree(self, tree):
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')
        conditional_module: Module = self.compiler_instance.get_action('CONDITIONAL')

        if_condition = tree['CONDITION']
        if_program = tree['PROGRAM']


        conditionals = {
            'EQEQ': conditional_module.equal,
            'NOT_EQEQ': conditional_module.nequal,
            'LESS': conditional_module.less,
            'GREATER': conditional_module.greater
        }


        # Resolve condition
        condition = conditionals[if_condition[0]]

        # Process first and second value
        first = resolution_module(if_condition[1])
        second = resolution_module(if_condition[2])
        
        def atomize(val):
            if type(val) == ID:
                val = self.compiler_instance.get_variable(val.value)['object']
            return val
        
        first_val = atomize(first)
        second_val = atomize(second)


        # Evaluate condition and run it's code
        while condition(first_val, second_val):
            self.compiler_instance.run(if_program)

            # Get vars again
            first_val = atomize(first)
            second_val = atomize(second)
    

class ForEachMod(Module):
    name="FOREACH"
    type = Module.MODULE_TYPES.ACTION

    def __init__(self, compiler_instance):
        super().__init__()

        self.compiler_instance = compiler_instance

    
    def proc_tree(self, tree):
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')

        iterable = resolution_module(tree['ITERABLE'])
        variable = resolution_module(tree['VARIABLE'])

        line = tree['ITERABLE'][-1]

        finished = []

        program = tree['PROGRAM']

        if type(iterable) == ID:
            var = self.compiler_instance.get_variable(iterable.value)

            if var['type'] == DynArray:
                for value in var['object'].value:
                    # Inject value as the variable
                    

                    instance = self.compiler_instance.new_instance(
                        tree=program,
                        sync="advanced"
                    )

                    instance.create_variable(
                        variable.value,
                        type(value),
                        value,
                        force = True
                    )

                    instance.run()

                    if self.no_construct:
                        finished.append(
                            ''.join(instance.finished)
                        )
                    else:
                        self.compiler_instance.write_back(
                            instance
                        )
                    self.compiler_instance.sync(instance)

            else:
                raise TranspilerExceptions.TypeMissmatch(f"Get ID Index Actor {var['object']}", var['type'], [ID, DynArray], line)
        
        elif type(iterable) == DynArray:
            raise TranspilerExceptions.NotImplemented
        
        else:
            raise TranspilerExceptions.TypeMissmatch("Get Index Actor", type(iterable), [ID, DynArray], line)
        
        if self.no_construct:
            return String(''.join(finished))

class InternalMod(Module):
    name="INTERNAL_CALL"
    type = Module.MODULE_TYPES.SPECIAL_ACTION

    def __init__(self, compiler_instance):
        super().__init__()

        self.compiler_instance = compiler_instance

    def panic(
        self,
        line
    ):
        raise RuntimeExceptions.Panic("Runtime panic. This panic was propagated by an explicit call.",
                                      line=line)
    
    def namespace(self, line):
        return print(self.compiler_instance.namespace)
    
    def debug(self, line):
        echo: Module = self.compiler_instance.get_action('echo')
        self.compiler_instance.finished.append(
f"""
 <table>
  <tr>
    <th>NoPHP Live Debug</th>
  </tr>
  <tr>
    <td>Current Namespace</td>
    <td>{self.compiler_instance.namespace}</td>
  </tr>
  <tr>
    <td></td>
    <td>Variables</td>
    {
        '<br>'.join(
            ["<tr><td>{}</td><td><pre><code>{}</code></pre></td></tr>".format(v, self.compiler_instance.get_variable(v)['object']) for v in self.compiler_instance.variables]
            )
    }
  </tr>
</table> 
"""
        )

    def die(self, line):
        self.compiler_instance.stop = True
    
    def proc_tree(self, tree):
        resolution_module: Module = self.compiler_instance.get_action('RESOLUT')

        value = resolution_module(tree['VALUE'])

        calls = {
            "panic": self.panic,
            "nm": self.namespace,
            "die": self.die,
            "debug": self.debug,
        }

        if value.value in calls:
            return calls[value.value](tree['VALUE'][-1])
        else:
            raise TranspilerExceptions.NotImplemented

        # Hardcode some features for now
        


"""
# This module handles all functions related to Strings
"""
from ..exceptions import TranspilerExceptions
from ..module import Module
from ..types import DynArray, String, BasicType, ID, Int32, sInnerMut

class StrMod(Module):
    def __init__(self, compiler_instance):
        super().__init__()
        self.compiler_instance = compiler_instance

    def base(self, tree, ref=False):
        values = []

        if 'POSITIONAL_ARGS' in tree['FUNCTION_ARGUMENTS']:
            for var in tree['FUNCTION_ARGUMENTS']['POSITIONAL_ARGS']:

                value = self.ref_resolve(var) if ref else self.safely_resolve(var)

                values.append(value)

        return values


class StrLenMod(StrMod):
    name = "strlen"
    type = Module.MODULE_TYPES.FUNCTION

    def __init__(self, compiler_instance):
        super().__init__(compiler_instance)

    def proc_tree(self, tree):
        values = self.base(tree)

        if len(values) != 1:
            raise TranspilerExceptions.IncorrectNumberOfValues(values, "strlen()")

        # Extracted the value from the list
        input_string = values[0]

        # Check if it's a string
        if not isinstance(input_string, str):
            raise TranspilerExceptions.TypeMissmatch("strlen()", "string", type(input_string).__name__, 0)

        # Calculate and return the length of the string
        length = len(input_string)
        return Int32(length)
    

class StrReplaceMod(StrMod):
    name = "str_replace"
    type = Module.MODULE_TYPES.FUNCTION

    def __init__(self, compiler_instance):
        super().__init__(compiler_instance)

    def proc_tree(self, tree):
        values = self.base(tree)

        if len(values) != 3:
            raise TranspilerExceptions.IncorrectNumberOfValues(values, "str_replace()")

        # Extracted the values from the list
        search_str, replace_str, subject_str = values

        # Check if all arguments are strings

        # Perform the string replacement
        result_str = subject_str.replace(search_str, replace_str)
        return String(result_str)
    
class Nl2BrMod(StrMod):
    name = "nl2br"
    type = Module.MODULE_TYPES.FUNCTION

    def __init__(self, compiler_instance):
        super().__init__(compiler_instance)

    def proc_tree(self, tree):
        values = self.base(tree)

        if len(values) != 1:
            raise TranspilerExceptions.IncorrectNumberOfValues(values, "nl2br($str)")

        # Extracted the values from the list
        subject_str = values[0]

        # Check if all arguments are strings

        # Perform the string replacement
        result_str = subject_str.replace('\n', '</br>')
        return String(result_str)

class SubstrMod(StrMod):
    name = "substr"
    type = Module.MODULE_TYPES.FUNCTION

    def __init__(self, compiler_instance):
        super().__init__(compiler_instance)

    def proc_tree(self, tree):
        values = self.base(tree)

        if len(values) < 2 or len(values) > 3:
            raise TranspilerExceptions.IncorrectNumberOfValues(values, "substr()")

        # Extract the values from the list
        subject_str, start, length = self.remove_quotes(values[0]), int(values[1]), None

        if len(values) == 3:
            length = int(values[2])

        # Check if all arguments are strings or integers
        if not isinstance(subject_str, str) or not isinstance(start, int) or (length is not None and not isinstance(length, int)):
            raise TranspilerExceptions.TypeMissmatch("substr()", "string, int[, int]", 0)

        # Perform the substring operation
        result_str = subject_str[start:start+length] if length is not None else subject_str[start:]
        # print("res:",result_str, subject_str, start, length)
        return String(result_str)


from sly import Parser
from sly.yacc import YaccProduction, YaccSymbol
import logging

from .lexer import PyettyLexer

class PyettyParser(Parser):
    tokens = PyettyLexer.tokens
    debugfile = "parser.out"
    log = logging.getLogger()
    log.setLevel(logging.ERROR)
    #syntax_error_obj = syntax_error()

    def parse(self, tokens, task = None, progress = None, tasklen=100):
        '''
        Parse the given input tokens.
        '''
        ERROR_COUNT = 0
        lookahead = None                                  # Current lookahead symbol
        lookaheadstack = []                               # Stack of lookahead symbols
        actions = self._lrtable.lr_action                 # Local reference to action table (to avoid lookup on self.)
        goto    = self._lrtable.lr_goto                   # Local reference to goto table (to avoid lookup on self.)
        prod    = self._grammar.Productions               # Local reference to production list (to avoid lookup on self.)
        defaulted_states = self._lrtable.defaulted_states # Local reference to defaulted states
        pslice  = YaccProduction(None)                    # Production object passed to grammar rules
        errorcount = 0                                    # Used during error recovery

        # Set up the state and symbol stacks
        self.tokens = tokens
        self.statestack = statestack = []                 # Stack of parsing states
        self.symstack = symstack = []                     # Stack of grammar symbols
        pslice._stack = symstack                          # Associate the stack with the production
        self.restart()

        # Set up position tracking
        track_positions = False
        if not hasattr(self, '_line_positions'):
            self._line_positions = { }           # id: -> lineno
            self._index_positions = { }          # id: -> (start, end)

        errtoken   = None                                 # Err token
        while True:
            # Update our task
            if task and progress:
                progress.update(task, advance=tasklen)

            # Get the next symbol on the input.  If a lookahead symbol
            # is already set, we just use that. Otherwise, we'll pull
            # the next token off of the lookaheadstack or from the lexer
            if self.state not in defaulted_states:
                if not lookahead:
                    if not lookaheadstack:
                        lookahead = next(tokens, None)  # Get the next token
                    else:
                        lookahead = lookaheadstack.pop()
                    if not lookahead:
                        lookahead = YaccSymbol()
                        lookahead.type = '$end'
                    
                # Check the action table
                ltype = lookahead.type
                t = actions[self.state].get(ltype)
            else:
                t = defaulted_states[self.state]

            if t is not None:
                if t > 0:
                    # shift a symbol on the stack
                    statestack.append(t)
                    self.state = t

                    symstack.append(lookahead)
                    lookahead = None

                    # Decrease error count on successful shift
                    if errorcount:
                        errorcount -= 1
                    continue

                if t < 0:
                    # reduce a symbol on the stack, emit a production
                    self.production = p = prod[-t]
                    pname = p.name
                    plen  = p.len
                    pslice._namemap = p.namemap

                    # Call the production function
                    pslice._slice = symstack[-plen:] if plen else []

                    sym = YaccSymbol()
                    sym.type = pname       
                    value = p.func(self, pslice)
                    if value is pslice:
                        value = (pname, *(s.value for s in pslice._slice))

                    sym.value = value
                        
                    # Record positions
                    if track_positions:
                        if plen:
                            sym.lineno = symstack[-plen].lineno
                            sym.index = symstack[-plen].index
                            sym.end = symstack[-1].end
                        else:
                            # A zero-length production  (what to put here?)
                            sym.lineno = None
                            sym.index = None
                            sym.end = None
                        self._line_positions[id(value)] = sym.lineno
                        self._index_positions[id(value)] = (sym.index, sym.end)
                            
                    if plen:
                        del symstack[-plen:]
                        del statestack[-plen:]

                    symstack.append(sym)
                    self.state = goto[statestack[-1]][pname]
                    statestack.append(self.state)
                    continue

                if t == 0:
                    n = symstack[-1]
                    result = getattr(n, 'value', None)
                    return result

            if t is None:
                # We have some kind of parsing error here.  To handle
                # this, we are going to push the current token onto
                # the tokenstack and replace it with an 'error' token.
                # If there are any synchronization rules, they may
                # catch it.
                #
                # In addition to pushing the error token, we call call
                # the user defined error() function if this is the
                # first syntax error.  This function is only called if
                # errorcount == 0.
                if errorcount == 0 or self.errorok:
                    errorcount = ERROR_COUNT
                    self.errorok = False
                    if lookahead.type == '$end':
                        errtoken = None               # End of file!
                    else:
                        errtoken = lookahead

                    tok = self.error(errtoken)
                    if tok:
                        # User must have done some kind of panic
                        # mode recovery on their own.  The
                        # returned token is the next lookahead
                        lookahead = tok
                        self.errorok = True
                        continue
                    else:
                        # If at EOF. We just return. Basically dead.
                        if not errtoken:
                            return
                else:
                    # Reset the error count.  Unsuccessful token shifted
                    errorcount = ERROR_COUNT

                # case 1:  the statestack only has 1 entry on it.  If we're in this state, the
                # entire parse has been rolled back and we're completely hosed.   The token is
                # discarded and we just keep going.

                if len(statestack) <= 1 and lookahead.type != '$end':
                    lookahead = None
                    self.state = 0
                    # Nuke the lookahead stack
                    del lookaheadstack[:]
                    continue

                # case 2: the statestack has a couple of entries on it, but we're
                # at the end of the file. nuke the top entry and generate an error token

                # Start nuking entries on the stack
                if lookahead.type == '$end':
                    # Whoa. We're really hosed here. Bail out
                    return

                if lookahead.type != 'error':
                    sym = symstack[-1]
                    if sym.type == 'error':
                        # Hmmm. Error is on top of stack, we'll just nuke input
                        # symbol and continue
                        lookahead = None
                        continue

                    # Create the error symbol for the first time and make it the new lookahead symbol
                    t = YaccSymbol()
                    t.type = 'error'

                    if hasattr(lookahead, 'lineno'):
                        t.lineno = lookahead.lineno
                    if hasattr(lookahead, 'index'):
                        t.index = lookahead.index
                    if hasattr(lookahead, 'end'):
                        t.end = lookahead.end
                    t.value = lookahead
                    lookaheadstack.append(lookahead)
                    lookahead = t
                else:
                    sym = symstack.pop()
                    statestack.pop()
                    self.state = statestack[-1]
                continue

            # Call an error function here
            raise RuntimeError('sly: internal parser error!!!\n')


    precedence = (
        ("left", EMPTY),
        ("left", ","),
        ("right", "="),
        ("left", "|"),
        ("left", "&"),
        ("left", EQEQ, NOT_EQEQ),
        ("left", EQ_LESS, EQ_GREATER, "<", ">"),
        ("left", "+", "-"),
        ("left", "*", "/", "%"),
        ("right", UMINUS, UPLUS),
        ("right", "!"),
        ("left", COLON_COLON),
    )

    # Program START
    @_("program statement")
    def program(self, p):
        return p.program + (p.statement,)

    @_("statement")
    def program(self, p):
        return (p.statement,)

    @_("empty")
    def program(self, p):
        return ()

    # Program END
    ###########################################################################
    # Statements START

    @_("function_declaration")
    def statement(self, p):
        return p.function_declaration + ()

    @_("class_declaration")
    def statement(self, p):
        return p.class_declaration

    @_("function_call_statement")
    def statement(self, p):
        return p.function_call_statement


    @_("conditional")
    def statement(self, p):
        return p.conditional

    @_("conditional")
    def expression(self, p):
        return p.conditional

    @_("while_loop")
    def statement(self, p):
        return p.while_loop

    @_("python_code_statement")
    def statement(self, p):
        return p.python_code_statement

    @_("variable_assignment")
    def statement(self, p):
        return p.variable_assignment

    @_("break_statement")
    def statement(self, p):
        return p.break_statement

    @_("for_loop")
    def statement(self, p):
        return p.for_loop
    
    @_("for_loop")
    def expression(self, p):
        return p.for_loop
    
    # @_("attribute")
    # def statement(self, p):
    #     return p.attribute

    @_("delete_statement")
    def statement(self, p):
        return p.delete_statement

    @_("return_statement")
    def statement(self, p):
        return p.return_statement

    @_("variable_operation")
    def statement(self, p):
        return p.variable_operation

    @_("import_statement")
    def statement(self, p):
        return p.import_statement

    @_("sandbox")
    def statement(self, p):
        return p.sandbox
    
    @_("expression")
    def statement(self, p):
        return p.expression
        

    def error(self, p):
        print("Whoa. You are seriously hosed.")
        if not p:
            print("End of File!")
        else:
            print("Failed at ", p)
        exit(1)

    # Statements END
    ###########################################################################
    # Statment syntax START

    @_("LIMPORT expression ';'")
    def sandbox(self, p):
        return ("LIMPORT", {"EXPRESSION": p.expression}, p.lineno)

    @_("PHPSTART program PHPEND")
    def sandbox(self, p):
        return ("PHP", {"PROGRAM": p.program}, p.lineno)
    

    @_("html_full")
    def expression(self, p):
        return p.html_full

    @_("HTMLSTART expression HTMLEND")
    def html_full(self, p):
        return ("HTML", {"PROGRAM": p.expression, "LABEL": {"START": p.HTMLSTART, "END": p.HTMLEND}}, p.lineno)

    @_("HTMLSTART empty HTMLEND")
    def html_full(self, p):
        return ("HTML", {"PROGRAM": [], "LABEL": {"START": p.HTMLSTART, "END": p.HTMLEND}}, p.lineno)
    
    ######################

    @_("html_full expression")
    def irregular_html_args(self, p):
        return p.irregular_html_args + (p.html_full,)
    
    #######################

    @_("HTMLSTART")
    def html_full(self, p):
        return ("HTML", {"PROGRAM": [], "LABEL": {"START": p.HTMLSTART, "END": None}}, p.lineno)
    
    @_("'?' HTMLEND")
    def html_full(self, p):
        return ("HTML", {"PROGRAM": [], "LABEL": {"START": None, "END": p.HTMLEND}}, p.lineno)

    @_("function_call ';'")
    def function_call_statement(self, p):
        return p.function_call

    @_("python_code ';'")
    def python_code_statement(self, p):
        return p.python_code

    @_("BREAK ';'")
    def break_statement(self, p):
        return ("BREAK", p.lineno)
    
    @_("SKIP ';'")
    def break_statement(self, p):
        return ("SKIP", p.lineno)
    
    @_("DEBUG ';'")
    def break_statement(self, p):
        return ("DEBUG", p.lineno)

    @_("RETURN expression ';'")
    def return_statement(self, p):
        return ("RETURN", {"EXPRESSION": p.expression}, p.lineno)
    
    
    
    @_("expression function_arguments")
    def function_call(self, p):
        return (
            "FUNCTION_CALL",
            {"FUNCTION_ARGUMENTS": p.function_arguments, "ID": p.expression},
        )

    @_("expression '(' function_arguments ')'")
    def function_call(self, p):
        return (
            "FUNCTION_CALL",
            {"FUNCTION_ARGUMENTS": p.function_arguments, "ID": p.expression},
        )


    @_("expression '$' ID '=' expression ';'")
    def function_call(self, p):
        return (
            "CLASS_ATTR",
            {"ID": p.ID, "VALUE": p.expression1, "LEVEL": p.expression0},
        )



    @_("expression '(' function_arguments ')' FARROW '{' program '}'")
    def function_call(self, p):
        return (
            "FUNCTION_CALL",
            {"FUNCTION_ARGUMENTS": p.function_arguments, "ID": p.expression,
             "ONCOMPLETE": p.program},
            p.lineno,
        )
    
    # $Router->myFunction();
    
    @_("'$' ID TARROW ID '(' function_arguments ')'")
    def function_call(self, p):
        return (
            "FUNCTION_CALL",
            {"FUNCTION_ARGUMENTS": p.function_arguments, "ID": p.ID1, "CLASS": p.ID0},
            p.lineno,
        )

    @_("'$' ID TARROW ID '(' empty ')'")
    def function_call(self, p):
        return (
            "FUNCTION_CALL",
            {"FUNCTION_ARGUMENTS": {}, "ID": p.ID1, "CLASS": p.ID0},
            p.lineno,
        )

    @_("ID COLON_COLON ID '(' function_arguments ')'")
    def function_call(self, p):
        return (
            "FUNCTION_CALL",
            {"FUNCTION_ARGUMENTS": p.function_arguments, "ID": p.ID1, "CLASS": p.ID0},
            p.lineno,
        )

    @_("ID COLON_COLON ID '(' empty ')'")
    def function_call(self, p):
        return (
            "FUNCTION_CALL",
            {"FUNCTION_ARGUMENTS": {}, "ID": p.ID1, "CLASS": p.ID0},
            p.lineno,
        )
    
    @_("NEW expression '(' function_arguments ')'")
    def function_call(self, p):
        return (
            "NEW_OBJECT",
            {"FUNCTION_ARGUMENTS": p.function_arguments, "ID": p.expression},
            p.lineno,
        )
    
    @_("NEW expression '(' empty ')'")
    def function_call(self, p):
        return (
            "NEW_OBJECT",
            {"FUNCTION_ARGUMENTS": {}, "ID": p.expression,},
            p.lineno,
        )
    
    @_("'?' expression ';'")
    def statement(self, p):
        return (
            "INTERNAL_CALL",
            {"VALUE": p.expression, "ID": p.expression},
            p.lineno,
        )

    @_("expression '(' empty ')'")
    def function_call(self, p):
        return (
            "FUNCTION_CALL",
            {"FUNCTION_ARGUMENTS": {}, "ID": p.expression},
            p.lineno,
        )
    
    @_("'#' DEFINE expression '?' ID")
    def function_call(self, p):
        return (
            "DEFINE",
            {
                "TO": p.ID, 
                "EXPRESSION": p.expression
            },
            p.lineno,
        )
    
    @_("'#' DEPENDS expression")
    def function_call(self, p):
        return (
            "DEPENDS",
            {
                "TO": p.expression
            },
            p.lineno,
        )
    
    @_("ID TARROW ID")
    def function_call(self, p):
        return (
            "TARROW",
            {
                "FROM": p.ID0, 
                "TO": p.ID1
            },
            p.lineno,
        )
    
    @_("ID '/' ID")
    def function_call(self, p):
        return (
            "NAMESPACE_MEMBER",
            {
                "NAMESPACE": p.ID0, 
                "MEMBER": p.ID1
            },
            p.lineno,
        )
    
    @_("ID '\\' ID")
    def function_call(self, p):
        return (
            "NAMESPACE_MEMBER",
            {
                "NAMESPACE": p.ID0, 
                "MEMBER": p.ID1
            },
            p.lineno,
        )
    
    @_("ID TARROW ID '(' function_arguments ')'")
    def function_call(self, p):
        return (
            "TARROW",
            {
                "FROM": p.ID0, 
                "TO": p.ID1,
                "ARGS": p.function_arguments
            },
            p.lineno,
        )

    @_("expression '(' empty ')' FARROW '{' program '}'")
    def function_call(self, p):
        return (
            "FUNCTION_CALL",
            {
                "FUNCTION_ARGUMENTS": {},
                "ID": p.expression,
                "ONCOMPLETE": p.program
            },
            p.lineno,
        )
    
    @_("'.' ENV '{' program '}'")
    def function_call(self, p):
        return (
            "FUNCTION_CALL",
            { 'FUNCTION_ARGUMENTS': {'POSITIONAL_ARGS': (('NULL', 'NULL'),)},
                "ID": ('ID', {'VALUE':'void'}),
                "ONCOMPLETE": p.program
            },
            p.lineno,
        )
    
    @_("'.' ENV FROM ID '{' program '}'")
    def function_call(self, p):
        return (
            "FUNCTION_CALL",
            { 'FUNCTION_ARGUMENTS': {'POSITIONAL_ARGS': (('NULL', 'NULL'),)},
                "ID": ('ID', {'VALUE':'void'}),
                "ONCOMPLETE": p.program,
                "ENV_FROM": p.ID
            },
            p.lineno,
        )

    @_("FUNC ID '(' function_arguments ')' ':' expression '{' program '}' ")
    def function_declaration(self, p):
        return (
            "FUNCTION_DECLARATION",
            {
                "FUNCTION_ARGUMENTS": p.function_arguments,
                "ID": p.ID,
                "PROGRAM": p.program,
                "RETURNS_TYPE": p.expression
            },
            p.lineno,
        )
    
    @_("FUNC ID '(' function_arguments ')' '{' program '}' ")
    def function_declaration(self, p):
        return (
            "FUNCTION_DECLARATION",
            {
                "FUNCTION_ARGUMENTS": p.function_arguments,
                "ID": p.ID,
                "PROGRAM": p.program
            },
            p.lineno,
        )


    @_("FUNC ID '(' empty ')' ':' expression '{' program '}'")
    def function_declaration(self, p):
        return (
            "FUNCTION_DECLARATION",
            {"FUNCTION_ARGUMENTS": {}, "ID": p.ID, "PROGRAM": p.program,
                "RETURNS_TYPE": p.expression},
            p.lineno,
        )
    
    @_("FUNC ID '(' empty ')'  '{' program '}'")
    def function_declaration(self, p):
        return (
            "FUNCTION_DECLARATION",
            {"FUNCTION_ARGUMENTS": {}, "ID": p.ID, "PROGRAM": p.program},
            p.lineno,
        )

    
    @_("ID FUNC ID '(' function_arguments ')' ':' expression '{' program '}' ")
    def function_declaration(self, p):
        return (
            "FUNCTION_DECLARATION",
            {
                "FUNCTION_ARGUMENTS": p.function_arguments,
                "ID": p.ID1,
                "LEVEL": p.ID0,
                "PROGRAM": p.program,
                "RETURNS_TYPE": p.expression
            },
            p.lineno,
        )
    
    @_("ID FUNC ID '(' empty ')' ':' expression '{' program '}' ")
    def function_declaration(self, p):
        return (
            "FUNCTION_DECLARATION",
            {
                "FUNCTION_ARGUMENTS": {},
                "ID": p.ID1,
                "LEVEL": p.ID0,
                "PROGRAM": p.program,
                "RETURNS_TYPE": p.expression
            },
            p.lineno,
        )
    
    @_("ID FUNC ID '(' function_arguments ')' '{' program '}' ")
    def function_declaration(self, p):
        return (
            "FUNCTION_DECLARATION",
            {
                "FUNCTION_ARGUMENTS": p.function_arguments,
                "ID": p.ID1,
                "LEVEL": p.ID0,
                "PROGRAM": p.program
            },
            p.lineno,
        )
    
    @_("ID FUNC ID '(' empty ')' '{' program '}' ")
    def function_declaration(self, p):
        return (
            "FUNCTION_DECLARATION",
            {
                "FUNCTION_ARGUMENTS": {},
                "ID": p.ID1,
                "LEVEL": p.ID0,
                "PROGRAM": p.program
            },
            p.lineno,
        )
    
    @_("INIT '(' function_arguments ')' '{' program '}' ';'")
    def function_declaration(self, p):
        return (
            "INIT",
            {
                "FUNCTION_ARGUMENTS": p.function_arguments,
                "ID": p.ID,
                "PROGRAM": p.program,
                "RETURNS_TYPE": p.expression
            },
            p.lineno,
        )

    @_("FUNC ID COLON_COLON ID '(' function_arguments ')' '{' program '}' TARROW expression")
    def function_declaration(self, p):
        return (
            "FUNCTION_DECLARATION",
            {
                "FUNCTION_ARGUMENTS": p.function_arguments,
                "NAMESPACE": p.ID0,
                "ID": p.ID1,
                "PROGRAM": p.program,
                "RETURNS_TYPE": p.expression
            },
            p.lineno,
        )
    
    @_("FUNC ID COLON_COLON ID '(' empty ')' '{' program '}' TARROW expression")
    def function_declaration(self, p):
        return (
            "FUNCTION_DECLARATION",
            {"FUNCTION_ARGUMENTS": {}, "ID": p.ID1, "PROGRAM": p.program, "NAMESPACE": p.ID0,
                "RETURNS_TYPE": p.expression},
            p.lineno,
        )

    @_("positional_args")
    def function_arguments(self, p):
        return {"POSITIONAL_ARGS": p.positional_args}
    

    @_("positional_args ',' kwargs")
    def function_arguments(self, p):
        return {"POSITIONAL_ARGS": p.positional_args, "KWARGS": p.kwargs}

    @_("kwargs")
    def function_arguments(self, p):
        return {"KWARGS": p.kwargs}

    @_("CLASS ID '{' program '}'")
    def class_declaration(self, p):
        return ("CLASS_DECLARATION", {"ID": p.ID, "PROGRAM": p.program}, p.lineno)

    @_("CLASS ID EXTENDS ID '{' program '}'")
    def class_declaration(self, p):
        return ("CLASS_DECLARATION", {"ID": p.ID0, "EXTENDS": p.ID1, "PROGRAM": p.program}, p.lineno)
    

    @_("NAMESPACE ID '{' program '}'")
    def class_declaration(self, p):
        return ("NAMESPACE", {"ID": p.ID, "PROGRAM": p.program}, p.lineno)
    
    @_("NAMESPACE ID ';'")
    def class_declaration(self, p):
        return ("NAMESPACE", {"ID": p.ID}, p.lineno)

    @_("FOR '(' expression ')' '{' program '}'")
    def for_loop(self, p):
        return (
            "FOR",
            {
                "PROGRAM": p.program,
                "VARIABLE": p.expression0,
                "ITERABLE": p.expression1,
            },
            p.lineno,
        )

    @_("FOREACH '(' expression AS expression ')' '{' program '}'")
    def for_loop(self, p):
        return (
            "FOREACH",
            {
                "ITERABLE": p.expression0,
                "VARIABLE": p.expression1,
                "PROGRAM": p.program,
            },
            p.lineno,
        )


    @_("WHILE '(' expression ')' '{' program '}'")
    def while_loop(self, p):
        return ("WHILE", {"PROGRAM": p.program, "CONDITION": p.expression}, p.lineno)

    @_("expression '.' expression")
    def expression(self, p):
        return ("CONCAT", {"0": p.expression0, "1": p.expression1})

    @_("positional_args ',' expression")
    def positional_args(self, p):
        return p.positional_args + (p.expression,)
    
    @_("expression")
    def positional_args(self, p):
        return (p.expression,)
    
    # @_("'$' expression")
    # def positional_args(self, p):
    #     return (p.expression,)
    

    @_("'$' expression")
    def expression(self, p):
        return p.expression

    @_("kwargs ',' id '=' expression")
    def kwargs(self, p):
        return p.kwargs + ({"ID": p.id, "EXPRESSION": p.expression},)

    @_("ID '=' expression")
    def kwargs(self, p):
        return ({"ID": p.ID, "EXPRESSION": p.expression},)

    @_("'$' ID '=' expression ';'")
    def variable_assignment(self, p):
        return (
            "VARIABLE_ASSIGNMENT",
            {"ID": p.ID, "EXPRESSION": p.expression},
            p.lineno,
        )
    
    @_("'$' ID TARROW ID '=' expression ';'")
    def variable_assignment(self, p):
        return (
            "VARIABLE_ASSIGNMENT",
            {"ID": p.ID1, "PARENT": p.ID0, "EXPRESSION": p.expression},
            p.lineno,
        )
    
    @_("LET ID ':' ID '=' expression ';'")
    def variable_assignment(self, p):
        return (
            "VARIABLE_ASSIGNMENT",
            {"ID": p.ID0, "TYPE":p.ID1, "EXPRESSION": p.expression},
            p.lineno,
        )

    @_("'$' get_index '=' expression ';'")
    def variable_assignment(self, p):
        return (
            "SET_INDEX",
            {"ID": p.get_index, "EXPRESSION": p.expression},
            p.lineno,
        )

    @_("ID EQ_ADD expression ';'")
    def variable_operation(self, p):
        return (
            "VARIABLE_OPERATION",
            {"ID": p.ID, "EXPRESSION": p.expression, "OPERATION": "ADD"},
            p.lineno,
        )

    @_("get_index EQ_ADD expression ';'")
    def variable_operation(self, p):
        return (
            "VARIABLE_OPERATION",
            {"ID": p.get_index, "EXPRESSION": p.expression, "OPERATION": "ADD"},
            p.lineno,
        )

    @_("ID EQ_SUB expression ';'")
    def variable_operation(self, p):
        return (
            "VARIABLE_OPERATION",
            {"ID": p.ID, "EXPRESSION": p.expression, "OPERATION": "SUB"},
            p.lineno,
        )

    @_("get_index EQ_SUB expression ';'")
    def variable_operation(self, p):
        return (
            "VARIABLE_OPERATION",
            {"ID": p.get_index, "EXPRESSION": p.expression, "OPERATION": "SUB"},
            p.lineno,
        )

    @_("ID EQ_MUL expression ';'")
    def variable_operation(self, p):
        return (
            "VARIABLE_OPERATION",
            {"ID": p.ID, "EXPRESSION": p.expression, "OPERATION": "MUL"},
            p.lineno,
        )

    @_("get_index EQ_MUL expression ';'")
    def variable_operation(self, p):
        return (
            "VARIABLE_OPERATION",
            {"ID": p.get_index, "EXPRESSION": p.expression, "OPERATION": "MUL"},
            p.lineno,
        )

    @_("ID EQ_MOD expression ';'")
    def variable_operation(self, p):
        return (
            "VARIABLE_OPERATION",
            {"ID": p.ID, "EXPRESSION": p.expression, "OPERATION": "MOD"},
            p.lineno,
        )

    @_("get_index EQ_MOD expression ';'")
    def variable_operation(self, p):
        return (
            "VARIABLE_OPERATION",
            {"ID": p.get_index, "EXPRESSION": p.expression, "OPERATION": "MOD"},
            p.lineno,
        )

    @_("ID EQ_DIV expression ';'")
    def variable_operation(self, p):
        return (
            "VARIABLE_OPERATION",
            {"ID": p.ID, "EXPRESSION": p.expression, "OPERATION": "DIV"},
            p.lineno,
        )

    @_("get_index EQ_DIV expression ';'")
    def variable_operation(self, p):
        return (
            "VARIABLE_OPERATION",
            {"ID": p.get_index, "EXPRESSION": p.expression, "OPERATION": "DIV"},
            p.lineno,
        )

    # @_("class_attribute '=' expression ';'")
    # def class_attribute_assignment(self, p):
    #     return (
    #         "CLASS_ATTRIBUTE_ASSIGNMENT",
    #         {"CLASS_ATTRIBUTE": p.class_attribute, "EXPRESSION": p.expression},
    #         p.lineno,
    #     )

    @_("if_statement")
    def conditional(self, p):
        return (
            "CONDITIONAL",
            {"IF": p.if_statement, "ELSE_IF": (
                None, None), "ELSE": (None, None)},
            p.if_statement[2],
        )

    @_("if_statement else_if_loop")
    def conditional(self, p):
        return (
            "CONDITIONAL",
            {"IF": p.if_statement, "ELSE_IF": p.else_if_loop,
                "ELSE": (None, None)},
            p.if_statement[2],
        )

    @_("if_statement else_if_loop else_statement")
    def conditional(self, p):
        return (
            "CONDITIONAL",
            {"IF": p.if_statement, "ELSE_IF": p.else_if_loop,
                "ELSE": p.else_statement},
            p.if_statement[2],
        )

    @_("if_statement else_statement")
    def conditional(self, p):
        return (
            "CONDITIONAL",
            {"IF": p.if_statement, "ELSE_IF": (
                None, None), "ELSE": p.else_statement},
            p.if_statement[2],
        )

    @_("IF '(' expression ')' '{' program '}'")
    def if_statement(self, p):
        return ("IF", {"CODE": p.program, "CONDITION": p.expression}, p.lineno)

    @_("else_if_loop else_if_statement")
    def else_if_loop(self, p):
        return p.else_if_loop + (p.else_if_statement,)

    @_("else_if_statement")
    def else_if_loop(self, p):
        return ("ELSE_IF", p.else_if_statement)

    @_("ELSE IF '(' expression ')' '{' program '}'")
    def else_if_statement(self, p):
        return ({"CODE": p.program, "CONDITION": p.expression}, p.lineno)

    @_("ELSE '{' program '}'")
    def else_statement(self, p):
        return ("ELSE", {"CODE": p.program}, p.lineno)

    @_("DEL ID ';'")
    def delete_statement(self, p):
        return ("DEL", {"ID": p.ID}, p.lineno)

    @_("IMPORT expression ';'")
    def import_statement(self, p):
        return ("IMPORT", {"EXPRESSION": p.expression}, p.lineno)

    @_("'.' GLOBAL ';'")
    def import_statement(self, p):
        return ("GLOBALS", {"VALUE":""}, p.lineno)
    
    @_("'.' SELFISH ';'")
    def import_statement(self, p):
        return ("SELFISH", {"VALUE":""}, p.lineno)

    # Statment syntax END
    ###########################################################################
    # Expression START

    @_("ID OF ID")
    def expression(self, p):
        return ("TYPED", {"ID": p.ID0, "TYPE": p.ID1}, p.lineno)
    
    @_("'&' ID")
    def expression(self, p):
        return ("POINTER", {"ID": p.ID}, p.lineno)
    
    @_("'*' ID")
    def expression(self, p):
        return ("RESOLVE", {"ID": p.ID}, p.lineno)
    
    @_("function_call")
    def expression(self, p):
        return p.function_call

    @_("expression '+' expression")
    def expression(self, p):
        return ("ADD", p[0], p[2])

    @_("expression '-' expression")
    def expression(self, p):
        return ("SUB", p[0], p[2])

    @_("expression '/' expression")
    def expression(self, p):
        return ("DIV", p[0], p[2])

    @_("expression '*' expression")
    def expression(self, p):
        return ("MUL", p[0], p[2])

    @_("expression '%' expression")
    def expression(self, p):
        return ("MOD", p[0], p[2])

    @_("expression EQEQ expression")
    def expression(self, p):
        return ("EQEQ", p[0], p[2])

    @_("expression NOT_EQEQ expression")
    def expression(self, p):
        return ("NOT_EQEQ", p[0], p[2])

    @_("expression EQ_LESS expression")
    def expression(self, p):
        return ("EQ_LESS", p[0], p[2])

    @_("expression EQ_GREATER expression")
    def expression(self, p):
        return ("EQ_GREATER", p[0], p[2])

    # This stuff is broken...

    @_("function_call '|' function_call")
    def expression(self, p):
        return ("OR", p[0], p[2])

    # @_("function_call '|' expression")
    # def expression(self, p):
    #     return ("OR", p[0], p[2])

    @_("expression '&' expression")
    def expression(self, p):
        return ("AND", p[0], p[2])
    

    @_("'-' expression %prec UMINUS")
    def expression(self, p):
        return ("NEG", p.expression)

    @_("'+' expression %prec UPLUS")
    def expression(self, p):
        return ("POS", p.expression)


    @_("'!' expression")
    def expression(self, p):
        return ("NOT", p.expression)

    @_("expression '<' expression")
    def expression(self, p):
        return ("LESS", p[0], p[2])

    @_("expression '>' expression")
    def expression(self, p):
        return ("GREATER", p[0], p[2])

    @_("'(' expression ')'")
    def expression(self, p):
        return p.expression

    @_("python_code")
    def expression(self, p):
        return p.python_code

    @_("get_index")
    def expression(self, p):
        return p.get_index

    @_("null")
    def expression(self, p):
        return p.null

    @_("int")
    def expression(self, p):
        return p.int

    @_("float")
    def expression(self, p):
        return p.float

    @_("bool")
    def expression(self, p):
        return p.bool

    @_("string")
    def expression(self, p):
        return p.string
    

    @_("id")
    def expression(self, p):
        return p.id

    @_("class_attribute")
    def expression(self, p):
        return p.class_attribute

    @_("_tuple")
    def expression(self, p):
        return p._tuple

    @_("_list")
    def expression(self, p):
        return p._list

    # @_("_numpy")
    # def expression(self, p):
    #     return p._numpy

    @_("assoc_array")
    def expression(self, p):
        return p.assoc_array

    # Expression END
    ###########################################################################
    # Intermediate expression START

    @_("NULL")
    def null(self, p):
        return ("NULL", "NULL")
    

    @_(r"'[' empty ']'")
    def assoc_array(self, p):
        return ("ARRAY", {"ITEMS": ()})
    
    @_(r"'{' empty '}'")
    def assoc_array(self, p):
        return ("MAP", {"ITEMS": ()})


    @_(r"'[' assoc_array_items ']'")
    def assoc_array(self, p):
        return ("ASSOC_ARRAY", {"ITEMS": p.assoc_array_items})

    @_("assoc_array_items ',' expression SARROW expression")
    def assoc_array_items(self, p):
        return p.assoc_array_items + ((p.expression0, p.expression1),)

    @_("expression SARROW expression")
    def assoc_array_items(self, p):
        return ((p.expression0, p.expression1),)


    @_("expression '[' expression ']'")
    def get_index(self, p):
        return ("GET_INDEX", {"EXPRESSION": p.expression0, "INDEX": p.expression1, "LINE": p.lineno}, p.lineno)

    @_("expression '^' expression")
    def get_index(self, p):
        return ("GET_ASSOC_INDEX", {"EXPRESSION": p.expression0, "INDEX": p.expression1}, p.lineno)


    @_("'{' positional_args '}'")
    def _tuple(self, p):
        return ("TUPLE", {"ITEMS": p.positional_args})

    @_("'{' positional_args ',' '}'")
    def _tuple(self, p):
        return ("TUPLE", {"ITEMS": p.positional_args})

    @_("'[' positional_args ']'")
    def _list(self, p):
        return ("ARRAY_FILLED", {"ITEMS": p.positional_args})

    @_("'[' positional_args ',' ']'")
    def _list(self, p):
        return ("ARRAY_FILLED", {"ITEMS": p.positional_args})

    # @_("'(' items ')'")
    # def _numpy(self, p):
    #     return ("ARRAY_FILLED", {"ITEMS": p.items})

    # @_("'(' items ',' ')'")
    # def _numpy(self, p):
    #     return ("ARRAY_FILLED", {"ITEMS": p.items})

    # @_("'(' expression ',' ')'")
    # def _numpy(self, p):
    #     return ("ARRAY_FILLED", {"ITEMS": (p.expression,)})

    # @_("'(' ')'")
    # def _numpy(self, p):
    #     return ("ARRAY_FILLED", {"ITEMS": ()})

    # @_("'(' ',' ')'")
    # def _numpy(self, p):
    #     return ("ARRAY_FILLED", {"ITEMS": ()})

    @_("items ',' expression")
    def items(self, p):
        return p.items + (p.expression,)

    # TODO: FIX THIS
    @_("expression ',' expression")
    def items(self, p):
        return (p.expression0,)

    @_("INT")
    def int(self, p):
        return ("INT", {"VALUE": p.INT})

    @_("CHAR")
    def string(self, p):
        return ("CHAR", {"VALUE": p.CHAR[1:-1]})

    @_("STRING")
    def string(self, p):
        return ("STRING", {"VALUE": p.STRING[1:-1]})

    @_("MULTILINE_STRING")
    def string(self, p):
        return ("STRING", {"VALUE": p.MULTILINE_STRING[1:-1]})
    
    @_("FORMATTED_STRING")
    def string(self, p):
        return ("FORMATTED_STRING", {"VALUE": p.FORMATTED_STRING[2:-1]})

    @_("FLOAT")
    def float(self, p):
        return ("FLOAT", {"VALUE": p.FLOAT})

    @_("TRUE")
    def bool(self, p):
        return ("BOOL", {"VALUE": p.TRUE})

    @_("FALSE")
    def bool(self, p):
        return ("BOOL", {"VALUE": p.FALSE})

    @_("expression COLON_COLON ID")
    def class_attribute(self, p):
        return ("CLASS_ATTRIBUTE", {"CLASS": p[0], "ATTRIBUTE": p[2]}, p.lineno)

    @_("ID")
    def id(self, p):
        return ("ID", {"VALUE": p.ID}, p.lineno)
    
    @_("PYTHON_CODE")
    def python_code(self, p):
        return ("PYTHON_CODE", {"CODE": p.PYTHON_CODE[2:-1]})

    @_("PYTHON_CODE_EXEC")
    def python_code(self, p):
        return ("PYTHON_CODE_EXEC", {"CODE": p.PYTHON_CODE_EXEC[3:-1]})

    @_("%prec EMPTY")
    def empty(self, p):
        pass

    # Intermediate expression END
    ###########################################################################
    # Syntax error START



lexer = PyettyLexer()
parser = PyettyParser()

def run(file):
    toks = []
    with open(file, "r") as f:
        text = f.read()
        toks = lexer.tokenize(text)
    ast = parser.parse(toks)
    return ast 